# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Antoine COLLET

"""Provide covariance matrix representation."""

from __future__ import annotations

import abc
import logging
from abc import abstractmethod
from functools import cached_property
from time import time
from typing import Callable, List, Optional, Sequence, Tuple, Union

import numpy as np
import scipy as sp
from numpy.random import Generator, RandomState
from scipy.sparse import csc_array, csr_array
from scipy.sparse.linalg import LinearOperator, gmres
from scipy.spatial.distance import cdist

from covmats._helpers import (
    check_random_state,
    get_pts_coords_regular_grid,
)
from covmats._sparse_helpers import (
    SparseCholeskyFactor,
)
from covmats._toeplitz import (
    create_toepliz_first_row,
    toeplitz_product,
)
from covmats._types import ArrayLike, NDArrayFloat, NDArrayInt


class CallBack:
    """Represents a callback instance."""

    __slots__: List[str] = ["res"]

    def __init__(self) -> None:
        """Initialize the instance."""
        self.res: List[NDArrayFloat] = []

    def __call__(self, rk) -> None:
        self.res.append(rk)

    @property
    def itercount(self) -> int:
        """Return the number of times the callback as been called."""
        return len(self.res)

    def clear(self) -> None:
        """Delete all results."""
        self.res = []


class CovarianceMatrix(LinearOperator, sp.stats.Covariance, abc.ABC):
    """
    Abstract representation of a covariance matrix.

    Calculations involving covariance matrices (e.g. data whitening,
    multivariate normal function evaluation) are often performed more
    efficiently using a decomposition of the covariance matrix instead of the
    covariance matrix itself. This class allows the user to construct an
    object representing a covariance matrix using any of several
    decompositions and perform calculations using a common interface.

    .. note::

        The `CovarianceMatrix` class cannot be instantiated directly. Instead, use
        one of the derived class:

        - :py:class:`CovViaDiagonal`
        - :py:class:`CovViaCholesky`
        - :py:class:`CovViaSparseCholesky`
        - :py:class:`CovViaPrecisionCholesky`
        - :py:class:`CovViaSparsePrecisionCholesky`
        - :py:class:`CovViaEigenFactorization`
        - :py:class:`CovViaEnsemble`


    Notes
    -----
    None of the :py:class:`Covariance` child implementation supports singular
    or indefinite covariance matrices. Otherwise, they would not be invertible,
    colorizing and whitening would not well defined.
    """

    __slots__ = [
        "_rank",
        "_n_pts",
        "_log_pdet",
        "_allow_singular",
        "_subspace_size",
    ]

    def __init__(self) -> None:
        """Initialize the instance."""
        # counters
        self.dtype = np.dtype("d")  # float64 for LinearOperator
        self._allow_singular = False  # must be invertible

    @property
    def n_pts(self) -> int:
        """Number of points in the domain (n)."""
        return self._n_pts

    @n_pts.setter
    def n_pts(self, n_pts: int) -> None:
        """Set the number of points in the domain (n)."""
        assert int(n_pts) > 0
        self._n_pts = n_pts

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the covariance matrix (n, n)."""
        return (self.n_pts, self.n_pts)

    @abstractmethod
    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        """Solve Ax = b, with A, the current covariance matrix instance."""

    @abstractmethod
    def get_diagonal(self) -> NDArrayFloat:
        """Return the diagonal entries of the matrix (variances) as a vector (n,)."""
        ...

    def get_trace(self) -> float:
        """Return the trace of the covariance matrix (sum of diagonal elements)."""
        return np.sum(self.get_diagonal()).item()

    @property
    def log_pdet(self) -> float:
        """
        Log of the pseudo-determinant of the covariance matrix.

        Note
        ----

        In linear algebra and statistics, the pseudo-determinant [1] is the product of
        all non-zero eigenvalues of a square matrix. It coincides with the regular
        determinant when the matrix is non-singular.
        """
        return np.array(self._log_pdet, dtype=float)[()]

    @property
    def rank(self) -> int:
        """
        Rank of the covariance matrix.
        """
        return np.array(self._rank, dtype=int)[()]

    @property
    def subspace_size(self) -> int:
        """
        Subspace size of the covariance matrix.
        """
        return self._subspace_size

    @abstractmethod
    def _todense(self) -> NDArrayFloat:
        """
        Return an explicit dense representation of the covariance matrix.
        """
        ...

    def todense(self) -> NDArrayFloat:
        """Explicit dense representation of the covariance matrix with shape (n, n)."""
        return self._todense()

    @property
    def covariance(self) -> NDArrayFloat:
        """
        Explicit dense representation of the covariance matrix.

        Notes
        -----
        Alias for :py:meth:`CovarianceMatrix.todense()`.
        """
        return self.todense()

    @property
    @abstractmethod
    def precision(self) -> NDArrayFloat:
        """
        Explicit dense representation of the precision matrix with shape (n, n).
        """
        ...

    @staticmethod
    def _validate_cov_mat(A: Union[NDArrayFloat, sp.sparse.sparray], name: str) -> None:
        """Raise an error if the input array is not square and of real number values."""
        m, n = A.shape[-2:]
        if (
            m != n
            or A.ndim != 2
            or not (
                np.issubdtype(A.dtype, np.integer)
                or np.issubdtype(A.dtype, np.floating)
            )
        ):
            message = (
                f"The input `{name}` must be a square, "
                "two-dimensional array of real numbers."
            )
            raise ValueError(message)

    @staticmethod
    def _validate_dense_matrix(A: ArrayLike, name: str) -> NDArrayFloat:
        """Raise an error if the input array is not square and of real number values."""
        A = np.atleast_2d(A)
        CovarianceMatrix._validate_cov_mat(A, name)
        return A

    @staticmethod
    def _validate_sparse_matrix(A: sp.sparse.sparray, name: str) -> sp.sparse.sparray:
        """Raise an error if the input array is not square and of real number values."""
        CovarianceMatrix._validate_cov_mat(A, name)
        return A

    @staticmethod
    def _validate_dense_lower_triangle(A: ArrayLike, name: str) -> NDArrayFloat:
        if not np.allclose(A, np.tril(A)):
            message = f"The input `{name}` must be a lower-triangular matrix."
            raise ValueError(message)
        return A

    @staticmethod
    def _validate_vector(A, name):
        """Raise an error if the input array is not a vector of real number values."""
        A = np.atleast_1d(A)
        if A.ndim != 1 or not (
            np.issubdtype(A.dtype, np.integer) or np.issubdtype(A.dtype, np.floating)
        ):
            message = (
                f"The input `{name}` must be a one-dimensional array of real numbers."
            )
            raise ValueError(message)
        return A

    def _rmatvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return A.T @ x."""
        return self._matvec(x)

    def _matmat(self, X: NDArrayFloat) -> NDArrayFloat:
        """Return A @ X."""
        return self._matvec(X)

    def _rmatmat(self, X: NDArrayFloat) -> NDArrayFloat:
        """Return A @ X."""
        return self._matmat(X)

    @abc.abstractmethod
    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """Perform a whitening transformation on data."""
        ...

    def whiten(self, x):
        """
        Perform a whitening transformation on data.

        "Whitening" ("white" as in "white noise", in which each frequency has
        equal magnitude) transforms a set of random variables into a new set of
        random variables with unit-diagonal covariance. When a whitening
        transform is applied to a sample of points distributed according to
        a multivariate normal distribution with zero mean, the covariance of
        the transformed sample is approximately the identity matrix
        :cite:p:`WhiteningTransformation2025, novakGeneralizationColoringLinear2019`.

        Parameters
        ----------
        x : array_like
            An array of points. The last dimension must correspond with the
            dimensionality of the space, i.e., the number of columns in the
            covariance matrix.

        Returns
        -------
        x_ : array_like
            The transformed array of points.

        References
        ----------
        .. bibliography::
            :list: enumerated
            :filter: False

            novakGeneralizationColoringLinear2019
            WhiteningTransformation2025

        Examples
        --------
        >>> import numpy as np
        >>> import covmats
        >>> rng = np.random.default_rng()
        >>> n = 3
        >>> A = rng.random(size=(n, n))
        >>> cov_array = A @ A.T  # make matrix symmetric positive definite
        >>> precision = np.linalg.inv(cov_array)
        >>> cov_object = covmats.CovViaPrecisionCholesky(
        .. sp.linalg.cholesky(precision, lower=True))
        >>> x = rng.multivariate_normal(np.zeros(n), cov_array, size=(10000))
        >>> x_ = cov_object.whiten(x)
        >>> np.cov(x_, rowvar=False)  # near-identity covariance
        array([[0.97862122, 0.00893147, 0.02430451],
               [0.00893147, 0.96719062, 0.02201312],
               [0.02430451, 0.02201312, 0.99206881]])

        """
        return self._whiten(np.asarray(x))

    @abc.abstractmethod
    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        """Perform a colorizing transformation on data."""
        ...

    def colorize(self, x):
        """
        Perform a colorizing transformation on data.

        "Colorizing" ("color" as in "colored noise", in which different
        frequencies may have different magnitudes) transforms a set of
        uncorrelated random variables into a new set of random variables with
        the desired covariance. When a coloring transform is applied to a
        sample of points distributed according to a multivariate normal
        distribution with identity covariance and zero mean, the covariance of
        the transformed sample is approximately the covariance matrix used
        in the coloring transform
        :cite:p:`WhiteningTransformation2025,novakGeneralizationColoringLinear2019`.

        Parameters
        ----------
        x : array_like
            An array of points. The last dimension must correspond with the
            dimensionality of the space, i.e., the number of columns in the
            covariance matrix.

        Returns
        -------
        x_ : array_like
            The transformed array of points.

        References
        ----------
        .. bibliography::
            :list: enumerated
            :filter: False

            WhiteningTransformation2025
            novakGeneralizationColoringLinear2019

        Examples
        --------
        >>> import numpy as np
        >>> import covmats
        >>> rng = np.random.default_rng(1638083107694713882823079058616272161)
        >>> n = 3
        >>> A = rng.random(size=(n, n))
        >>> cov_array = A @ A.T  # make matrix symmetric positive definite
        >>> cholesky = np.linalg.cholesky(cov_array)
        >>> cov_object = covmats.CovViaCholesky(cholesky)
        >>> x = rng.multivariate_normal(np.zeros(n), np.eye(n), size=(10000))
        >>> x_ = cov_object.colorize(x)
        >>> cov_data = np.cov(x_, rowvar=False)
        >>> np.allclose(cov_data, cov_array, rtol=3e-2)
        True
        """
        return self._colorize(np.asarray(x))

    def sample_mvnormal(
        self,
        shape: Sequence[int],
        random_state: Optional[Union[int, RandomState, Generator]] = None,
    ) -> NDArrayFloat:
        r"""
        Draw samples from the multivariate normal N(0, A).

        Parameters
        ----------
        shape: Sequence[int]
            Number of random vectors to sample. The resulting array will be of shape
            (shape, n), n being the number of elements per random vector
            (the covariance) matrix has shape (n, n).
        random_state: Optional[Union[int, np.random.Generator, np.random.RandomState]]
            Pseudorandom number generator state used to generate resamples.
            If `random_state` is ``None`` (or `np.random`), the
            `numpy.random.RandomState` singleton is used.
            If `random_state` is an int, a new ``RandomState`` instance is used,
            seeded with `random_state`.
            If `random_state` is already a ``Generator`` or ``RandomState``
            instance then that instance is used. The default is None.

        Return
        ------
        X: The transformed array of points. It has shape (input shape, n), n being
        the number of elements per random vector (the covariance)
        matrix has shape (n, n).

        Examples
        --------
        >>> import numpy as np
        >>> import scipy as sp
        >>> import covmats
        >>> covd = covmats.CovViaDiagonal(np.array([5.0, 10.0, 15.0]))
        >>> rng_seed = 42
        >>> covd.sample_mvnormal(shape=[2], random_state=rng_seed)
        array([[ 1.11068661, -0.43723011,  2.50848692],
            [ 3.40559829, -0.74045799, -0.90680853]])
        >>> x = covd.sample_mvnormal(shape=[2, 4], random_state=rng_seed)
        >>> x
        array([[[ 1.11068661, -0.43723011,  2.50848692],
                [ 3.40559829, -0.74045799, -0.90680853],
                [ 3.53122721,  2.4268417 , -1.81826648],
                [ 1.21320114, -1.46545542, -1.80376358]],
               [[ 0.54104409, -6.05032338, -6.68057804],
                [-1.25731314, -3.20285323,  1.21707469],
                [-2.03040356, -4.46609644,  5.67643327],
                [-0.50485116,  0.21354293, -5.518026  ]]])
        >>> x.shape
        (2, 4, 3)
        >>> cov_cho = covmats.CovViaCholesky(sp.linalg.cholesky(covd.todense()))
        >>> cov_cho.sample_mvnormal(shape=[2], random_state=rng_seed)
        array([[-0.95777013, -1.11354406,  2.06162461],
            [ 0.81715777,  1.30517512,  1.66856257]])
        >>> cov_cho.sample_mvnormal(shape=[2, 2], random_state=rng_seed)
        array([[[ 1.11068661, -0.43723011,  2.50848692],
                [ 3.40559829, -0.74045799, -0.90680853],
                [ 3.53122721,  2.4268417 , -1.81826648],
                [ 1.21320114, -1.46545542, -1.80376358]],
               [[ 0.54104409, -6.05032338, -6.68057804],
                [-1.25731314, -3.20285323,  1.21707469],
                [-2.03040356, -4.46609644,  5.67643327],
                [-0.50485116,  0.21354293, -5.518026  ]]])

        """
        # A 1D diagonal of a covariance matrix was passed
        return self._colorize(
            check_random_state(seed=random_state).standard_normal(
                size=(*shape, self.subspace_size)
            )
        )

    @classmethod
    def from_cholesky(self, *args):
        """Representation of a covariance provided via choleksy factorization.

        .. error::
            Removed from parent class, do not use.
        """
        raise NotImplementedError(
            "`from_cholesky` is not available, please instantiate"
            " with `CovViaCholesky(...)` directly!"
        )

    @classmethod
    def from_diagonal(self, *args):
        """Representation of a covariance provided via diagonal.

        .. error::
            Removed from parent class, do not use.
        """
        raise NotImplementedError(
            "`from_diagonal` is not available, please instantiate"
            " with `CovViaDiagonal(...)` directly!"
        )

    @classmethod
    def from_eigendecomposition(self, *args):
        """
        Representation of a covariance provided via eigendecomposition.

        .. error::
            Removed from parent class, do not use.
        """
        raise NotImplementedError(
            "`from_eigendecomposition` is not available, please instantiate"
            " with `CovViaEigenFactorization(...)` directly!"
        )

    @classmethod
    def from_precision(self, *args):
        """
        Return a representation of a covariance from its precision matrix.

        .. error::
            Removed from parent class, do not use.
        """
        raise NotImplementedError(
            "`from_precision` is not available, please instantiate"
            " with `CovViaPrecisionCholesky(...)` directly!"
        )


def _dot_diag(x: NDArrayFloat, d: NDArrayFloat):
    """
    Perform the dot product between x with shape (..., n) and d with shape (n).

    Parameters
    ----------
    x : NDArrayFloat
        Input array with shape (..., n).
    d : NDArrayFloat
        1D vector representing a the diagonal elements of a (n, n) diagonal matrix.

    Returns
    -------
    A @ x
    """
    # If d were a full diagonal matrix, x @ d would always do what we want.
    # Special treatment is needed for n-dimensional `d` in which each row
    # includes only the diagonal elements of a covariance matrix.
    return x * d if x.ndim < 2 else x * np.expand_dims(d, -2)


class CovViaDiagonal(CovarianceMatrix):
    r"""
    Representation of a covariance matrix via its diagonal.

    Attributes
    ----------
    diagonal : NDArrayFloat
        The diagonal elements of a diagonal matrix.

    Notes
    -----
    Let the diagonal elements of a diagonal covariance matrix :math:`D` be
    stored in the vector :math:`d`.

    When all elements of :math:`d` are strictly positive, whitening of a
    data point :math:`x` is performed by computing
    :math:`x \cdot d^{-1/2}`, where the inverse square root can be taken
    element-wise.
    :math:`\log\det{D}` is calculated as :math:`-2 \sum(\log{d})`,
    where the :math:`\log` operation is performed element-wise.


    Notes
    -----
    No null nor negative elements are allowed, otherwise the matrix would be
    singular or indefinite and consequently non invertible.


    Examples
    --------
    Prepare a symmetric positive definite covariance matrix ``A`` and a
    data point ``x``.

    >>> import numpy as np
    >>> import covmats
    >>> rng = np.random.default_rng()
    >>> n = 5
    >>> A = np.diag(rng.random(n))
    >>> x = rng.random(size=n)

    Extract the diagonal from ``A`` and create the `Covariance` object.

    >>> d = np.diag(A)
    >>> cov = covmats.CovViaDiagonal(d)

    Compare the functionality of the `Covariance` object against a
    reference implementations.

    >>> res = cov.whiten(x)
    >>> ref = np.diag(d**-0.5) @ x
    >>> np.allclose(res, ref)
    True
    >>> res = cov.log_pdet
    >>> ref = np.linalg.slogdet(A)[-1]
    >>> np.allclose(res, ref)
    True

    """

    __slots__ = ["_D", "_sqrtD", "_invD", "_i_zero"]

    def __init__(self, diagonal: ArrayLike) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        diagonal : ArrayLike
            The diagonal elements of a diagonal matrix.
        """
        self.D = diagonal
        super().__init__()

    @property
    def D(self) -> NDArrayFloat:
        """1D vector representing the diagonal elements of the covariance matrix."""
        return self._D

    @D.setter
    def D(self, D: NDArrayFloat) -> None:
        """
        Set the diagonal elements of the covariance matrix.

        Notes
        -----
        No null nor negative elements are allowed, otherwise the matrix would be
        singular or indefinite and consequently non invertible.
        """
        self._D = self._validate_vector(
            A=np.asarray(D, dtype=np.float64), name="diagonal"
        )
        nnv = np.count_nonzero(self.D < 0.0)
        if nnv != 0:
            raise ValueError(
                f"{nnv} negative values have been detected in `D` which "
                "means the matrix is indefinite and non invertible."
            )

        nnv = np.count_nonzero(self.D == 0.0)
        if nnv != 0:
            raise ValueError(
                f"{nnv} null values have been detected in `D` which "
                "means the matrix is singular and non invertible."
            )
        self.n_pts = np.size(self.D)
        self._subspace_size = self.n_pts
        self._rank = self.n_pts
        # Update the pseudo inverse as well
        self._invD = 1.0 / self.D
        # Store the square roots for whitening and colorizing transformations
        self._sqrtD = np.sqrt(self.D)
        self._sqrtinvD = np.sqrt(self._invD)
        # Update the log pseudo determinant
        self._log_pdet = np.sum(np.log(self.D), axis=-1)

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        return self.get_diagonal() * x

    def _matmat(self, X: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the matrix X."""
        return self.get_diagonal()[:, np.newaxis] * X

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return _dot_diag(x, self._sqrtinvD)

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a colorizing transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to colorize with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return _dot_diag(x, self._sqrtD)

    def _todense(self) -> NDArrayFloat:
        return np.apply_along_axis(np.diag, -1, self.D)

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        """Solve Ax = b, with A, the current covariance matrix instance."""

        return self._invD * b if b.ndim < 2 else (self._invD)[:, np.newaxis] * b

    def get_diagonal(self) -> NDArrayFloat:
        """Return the diagonal entries of the matrix (variances)."""
        return self.D

    @property
    def precision(self) -> NDArrayFloat:
        """
        Explicit dense representation of the precision matrix.
        """
        return np.diag(self._invD)


class CovViaCholesky(CovarianceMatrix):
    r"""
    Representation of a covariance via a Cholesky factorization.

    Parameters
    ----------
    cholesky : array_like
        The lower triangular Cholesky factor of the covariance matrix.

    Notes
    -----
    Let the covariance matrix be :math:`A` and :math:`L` be the lower
    Cholesky factor such that :math:`L L^T = A`.
    Whitening of a data point :math:`x` is performed by computing
    :math:`L^{-1} x`. :math:`\log\det{A}` is calculated as
    :math:`2tr(\log{L})`, where the :math:`\log` operation is performed
    element-wise.

    This `Covariance` class does not support singular covariance matrices
    because the Cholesky decomposition does not exist for a singular
    covariance matrix.

    Examples
    --------
    Prepare a symmetric positive definite covariance matrix ``A`` and a
    data point ``x``.

    >>> import numpy as np
    >>> import covmats
    >>> rng = np.random.default_rng()
    >>> n = 5
    >>> A = rng.random(size=(n, n))
    >>> A = A @ A.T  # make the covariance symmetric positive definite
    >>> x = rng.random(size=n)

    Perform the Cholesky decomposition of ``A`` and create the
    `Covariance` object.

    >>> L = np.linalg.cholesky(A)
    >>> cov = covmats.CovViaCholesky(L)

    Compare the functionality of the `Covariance` object against
    reference implementation.

    >>> from scipy.linalg import solve_triangular
    >>> res = cov.whiten(x)
    >>> ref = solve_triangular(L, x, lower=True)
    >>> np.allclose(res, ref)
    True
    >>> res = cov.log_pdet
    >>> ref = np.linalg.slogdet(A)[-1]
    >>> np.allclose(res, ref)
    True

    """

    __slots__ = ["_L", "_covariance"]

    def __init__(self, L: ArrayLike, covariance: Optional[ArrayLike] = None) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        L : NDArrayFloat
            Lower triangle of the covariance matrix Cholesky factorization.
        covariance : Optional[NDArrayFloat], optional
            Dense covariance matrix, by default None
        """
        self.L = L
        super().__init__()
        if covariance is not None:
            self._covariance = self._validate_dense_matrix(covariance, "covariance")
        else:
            self._covariance = None

    @property
    def L(self) -> NDArrayFloat:
        """Lower triangle of the covariance matrix Cholesky factorization."""
        return self._L

    @L.setter
    def L(self, L: NDArrayFloat) -> None:
        """Set the lower triangle of the covariance matrix Cholesky factorization."""
        self._L = self._validate_dense_lower_triangle(
            self._validate_matrix(L, "L"), "L"
        )
        self._log_pdet = 2 * np.log(np.diag(self._L)).sum(axis=-1)
        self.n_pts = np.shape(self.L)[0]
        self._subspace_size = self.n_pts
        self._rank = self.n_pts

    @property
    def precision(self) -> NDArrayFloat:
        """
        Explicit dense representation of the precision matrix.
        """
        return self.solve(np.eye(self.n_pts))

    def _todense(self) -> NDArrayFloat:
        if self._covariance is not None:
            return self._covariance
        return self.L @ self.L.T

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        return np.linalg.multi_dot([self.L, self.L.T, x])

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        m = x.T.shape[0]
        res = sp.linalg.solve_triangular(self.L, x.T.reshape(m, -1), lower=True)
        return res.reshape(x.T.shape).T

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        return x @ self.L.T

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        """Solve Ax = b, with A, the current covariance matrix instance."""
        return sp.linalg.cho_solve((self.L, True), b)

    def get_diagonal(self) -> NDArrayFloat:
        """Return the diagonal entries of the matrix (variances)."""
        return np.sum((self.L**2), axis=1)


class CovViaSparseCholesky(CovarianceMatrix):
    r"""
    Representation of a covariance via a sparse Cholesky factorization.

    Notes
    -----
    Let the covariance matrix be :math:`A` and :math:`L` be the lower
    Cholesky factor such that :math:`L L^T = A`.
    Whitening of a data point :math:`x` is performed by computing
    :math:`L^{-1} x`. :math:`\log\det{A}` is calculated as
    :math:`2tr(\log{L})`, where the :math:`\log` operation is performed
    element-wise.

    This `Covariance` class does not support singular covariance matrices
    because the Cholesky decomposition does not exist for a singular
    covariance matrix.

    TODO: indicate that LGPL licence is expected here because it relies on CHOLMOD.

    Examples
    --------
    Prepare a symmetric positive definite covariance matrix ``A`` and a
    data point ``x``.

    >>> import numpy as np
    >>> import covmats
    >>> rng = np.random.default_rng()
    >>> n = 5
    >>> A = rng.random(size=(n, n))
    >>> A = A @ A.T  # make the covariance symmetric positive definite
    >>> x = rng.random(size=n)

    Perform the Cholesky decomposition of ``A`` and create the
    `Covariance` object.

    >>> L = np.linalg.cholesky(A)
    >>> cov = covmats.CovViaCholesky(L)

    Compare the functionality of the `Covariance` object against
    reference implementation.

    >>> from scipy.linalg import solve_triangular
    >>> res = cov.whiten(x)
    >>> ref = solve_triangular(L, x, lower=True)
    >>> np.allclose(res, ref)
    True
    >>> res = cov.log_pdet
    >>> ref = np.linalg.slogdet(A)[-1]
    >>> np.allclose(res, ref)
    True

    """

    __slots__ = ["_scf", "_sparse_covariance"]

    def __init__(
        self,
        scf: SparseCholeskyFactor,
        sparse_covariance: Optional[sp.sparse.sparray] = None,
    ) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        scf : SparseCholeskyFactor
            Lower triangle of the sparse covariance matrix Cholesky factorization.
        sparse_covariance : Optional[sp.sparse.sparray], optional
            Sparse covariance matrix, by default None
        """
        self.scf = scf
        if sparse_covariance is not None:
            self._validate_sparse_matrix(sparse_covariance, name="sparse_covariance")
        self._sparse_covariance: Optional[sp.sparse.sparray] = sparse_covariance
        super().__init__()

    @property
    def scf(self) -> SparseCholeskyFactor:
        return self._scf

    @scf.setter
    def scf(self, scf: SparseCholeskyFactor) -> None:
        self._scf = scf
        self._n_pts = self._scf.L.shape[0]
        self._log_pdet = self._scf.log_pdet
        self._subspace_size = self.n_pts
        self._rank = self._n_pts

    def _todense(self):
        return self.scf.todense()

    @property
    def precision(self) -> NDArrayFloat:
        """
        Explicit dense representation of the precision matrix.
        """
        return self.scf.inv()

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        if self._sparse_covariance is not None:
            return self._sparse_covariance @ x
        return self.scf @ x

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return self.scf.whiten(x)

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        return self.scf.colorize(x)

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        """Solve Ax = b, with A, the current covariance matrix instance."""
        return self.scf.solve(b)

    def get_diagonal(self) -> NDArrayFloat:
        """
        Return the diagonal entries of the matrix (variances).
        The matrix is never built explicitly. Instead the matvec interface is
        used to multiply all column of the identity matrix.
        """
        return self.scf.get_diagonal()


class CovViaPrecisionCholesky(CovarianceMatrix):
    r"""
    Representation of a covariance via the Cholesky factorization of its inverse
    (aka the precision matrix).

    Notes
    -----
    Let the covariance matrix be :math:`A`, its precision matrix be
    :math:`P = A^{-1}`, and :math:`L` be the lower Cholesky factor such
    that :math:`L L^T = P`.
    Whitening of a data point :math:`x` is performed by computing
    :math:`x^T L`. :math:`\log\det{A}` is calculated as
    :math:`-2tr(\log{L})`, where the :math:`\log` operation is performed
    element-wise.

    This `Covariance` class does not support singular covariance matrices
    because the precision matrix does not exist for a singular covariance
    matrix.

    Examples
    --------
    Prepare a symmetric positive definite precision matrix ``P`` and a
    data point ``x``. (If the precision matrix is not already available,
    consider the other factory methods of the `Covariance` class.)

    >>> import numpy as np
    >>> import covmats
    >>> rng = np.random.default_rng()
    >>> n = 5
    >>> P = rng.random(size=(n, n))
    >>> P = P @ P.T  # a precision matrix must be positive definite
    >>> x = rng.random(size=n)

    Create the `Covariance` object.

    >>> cov = covmats.CovViaPrecisionCholesky(np.linalg.cholesky(P))

    Compare the functionality of the `Covariance` object against
    reference implementations.

    >>> res = cov.whiten(x)
    >>> ref = x @ np.linalg.cholesky(P)
    >>> np.allclose(res, ref)
    True
    >>> res = cov.log_pdet
    >>> ref = -np.linalg.slogdet(P)[-1]
    >>> np.allclose(res, ref)
    True

    """

    __slots__: List[str] = ["_L", "_precision"]

    def __init__(self, L: ArrayLike, precision: Optional[ArrayLike] = None) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        L : NDArrayFloat
            Lower triangle of the precision matrix Cholesky factorization.
        covariance : Optional[NDArrayFloat], optional
            Dense precision matrix, by default None

        Raises
        ------
        ValueError
            _description_
        """
        self.L = L
        super().__init__()
        if precision is not None:
            self._precision: Optional[NDArrayFloat] = self._validate_dense_matrix(
                precision, "precision"
            )
        else:
            self._precision = None

    @property
    def L(self) -> NDArrayFloat:
        """Lower triangle of the precision matrix Cholesky factorization."""
        return self._L

    @L.setter
    def L(self, L: NDArrayFloat) -> None:
        """Set the lower triangle of the precision matrix Cholesky factorization."""
        self._L = self._validate_dense_lower_triangle(
            self._validate_matrix(L, "L"), "L"
        )
        self._log_pdet = -2 * np.log(np.diag(self.L)).sum(axis=-1)
        self.n_pts = np.shape(self.L)[0]
        self._subspace_size = self.n_pts
        self._rank = self.n_pts

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        return sp.linalg.cho_solve((self.L, True), x)

    def _todense(self) -> NDArrayFloat:
        return sp.linalg.cho_solve((self.L, True), np.eye(self.n_pts))

    @property
    def precision(self) -> NDArrayFloat:
        """
        Explicit dense representation of the precision matrix.
        """
        if self._precision is not None:
            return self._precision
        return self.L @ self.L.T

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return x @ self.L

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        m = x.T.shape[0]
        res = sp.linalg.solve_triangular(self.L.T, x.T.reshape(m, -1), lower=False)
        # L1^T @ b.T = x.T
        return res.reshape(x.T.shape).T

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        """Solve Ax = b, with A, the current covariance matrix instance."""
        if self._precision is not None:
            return self._precision @ b
        return self.L @ (self.L.T) @ b

    def get_diagonal(self) -> NDArrayFloat:
        """Return the diagonal entries of the matrix (variances)."""
        # Otherwise extract it using matrix vector products
        approx_diag = np.zeros(self.n_pts)
        # TODO: multiprocessing to speed up things ? Or by chunks of 1000 ?
        for i in range(self.n_pts):
            # construct the ith row of the identity matrix
            v = np.zeros(self.n_pts)
            v[i] = 1.0
            approx_diag[i] = self.matvec(v)[i]
        return approx_diag


class CovViaSparsePrecisionCholesky(CovarianceMatrix):
    """
    Representation of a covariance via the sparse Cholesky factorization of its
    sparse inverse (aka the precision matrix).

    Notes
    -----
    Blablabla.
    """

    __slots__: List[str] = [
        "_sparse_precision",
        "_scf",
    ]

    def __init__(
        self,
        scf: SparseCholeskyFactor,
        sparse_precision: Optional[sp.sparse.sparray] = None,
    ) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        scf : SparseCholeskyFactor
            Lower triangle of the sparse precision matrix Cholesky factorization.
        sparse_precision : Optional[sp.sparse.sparray], optional
            Sparse precision matrix (inverse of the covariance matrix), by default None
        """
        self.scf = scf

        if sparse_precision is not None:
            self._sparse_precision = self._validate_sparse_matrix(
                sparse_precision, "sparse_precision"
            )
        self._sparse_precision: Optional[sp.sparse.sparray] = sparse_precision
        super().__init__()

    @property
    def scf(self) -> SparseCholeskyFactor:
        return self._scf

    @scf.setter
    def scf(self, scf: SparseCholeskyFactor) -> None:
        self._scf = scf
        self._n_pts = self._scf.L.shape[0]
        self._log_pdet = -self._scf.log_pdet
        self._subspace_size = self.n_pts
        self._rank = self._n_pts

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        return self.scf.solve(x)

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return self.scf.whiten_inv(x)

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        return self.scf.colorize_inv(x)

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        """Return x = A^{-1} b."""
        if self._sparse_precision is not None:
            return self._sparse_precision.dot(b)
        return self.scf.mat @ b

    def get_diagonal(self) -> NDArrayFloat:
        """
        Return the diagonal entries of the matrix (variances).
        The matrix is never built explicitly. Instead the matvec interface is
        used to multiply all column of the identity matrix.
        """
        return self.scf.get_invdiagonal()

    def _todense(self):
        return self.scf.inv()

    @property
    def precision(self) -> sp.sparse.sparray:
        if self._sparse_precision is not None:
            return self._sparse_precision
        return self.scf.mat


class CovViaEigenFactorization(CovarianceMatrix):
    r"""
    Representation of a covariance provided via eigenfactorization

    Parameters
    ----------
    eigenfactorization : sequence
        A sequence (nominally a tuple) containing the eigenvalue and
        eigenvector arrays as computed by `scipy.sparse.eigsh`.

    Notes
    -----
    Let the covariance matrix be :math:`A`, let :math:`V` be matrix of
    eigenvectors, and let :math:`W` be the diagonal matrix of eigenvalues
    such that `V W V^T = A`.

    When all of the eigenvalues are strictly positive, whitening of a
    data point :math:`x` is performed by computing
    :math:`x^T (V W^{-1/2})`, where the inverse square root can be taken
    element-wise.
    :math:`\log\det{A}` is calculated as  :math:`tr(\log{W})`,
    where the :math:`\log` operation is performed element-wise.

    This `Covariance` class supports singular covariance matrices. When
    computing ``_log_pdet``, non-positive eigenvalues are ignored.
    Whitening is not well defined when the point to be whitened
    does not lie in the span of the columns of the covariance matrix. The
    convention taken here is to treat the inverse square root of
    non-positive eigenvalues as zeros.

    Examples
    --------
    Prepare a symmetric positive definite covariance matrix ``A`` and a
    data point ``x``.

    >>> import numpy as np
    >>> import covmats
    >>> rng = np.random.default_rng()
    >>> n = 5
    >>> A = rng.random(size=(n, n))
    >>> A = A @ A.T  # make the covariance symmetric positive definite
    >>> x = rng.random(size=n)

    Perform the eigenfactorization of ``A`` and create the `Covariance`
    object.

    >>> w, v = np.linalg.eigh(A)
    >>> cov = covmats.CovViaEigenFactorization((w, v))

    Compare the functionality of the `Covariance` object against
    reference implementations.

    >>> res = cov.whiten(x)
    >>> ref = x @ (v @ np.diag(w**-0.5))
    >>> np.allclose(res, ref)
    True
    >>> res = cov.log_pdet
    >>> ref = np.linalg.slogdet(A)[-1]
    >>> np.allclose(res, ref)
    True

    """

    __slots__: List[str] = ["_w", "_v", "_inv_w", "_sqrt_w", "_inv_sqrt_w"]

    def __init__(self, eigenfactorization: Tuple[NDArrayFloat, NDArrayFloat]) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        (eig_vals, eig_vects) : Tuple[NDArrayFloat, NDArrayFloat]
            - 1D vector of eigen values with size `n_pc`.
            - 2D arrays of eigen vectors (columns) with size `(Ns, n_pc)`. Ns being the
               number of elements in the original covariance matrix.

        """
        eigenvalues, eigenvectors = eigenfactorization
        # See if we do a common setter interface ?
        self.eig_vals = eigenvalues
        self.eig_vects = eigenvectors
        super().__init__()

    @property
    def n_pc(self) -> int:
        """
        Return the number of eigen vectors/values, i.e. principal components.

        It is determined from the eigen values vector size.
        """
        return self._w.size

    @property
    def subspace_size(self) -> int:
        return self.n_pc

    @property
    def eig_vals(self) -> NDArrayFloat:
        """Return the Eigen values."""
        return self._w

    @eig_vals.setter
    def eig_vals(self, value: NDArrayFloat) -> None:
        """Return the Eigen values."""
        self._w = np.reshape(value, (-1, 1))
        self._sqrt_w = np.sqrt(self._w)
        self._inv_w = 1.0 / self._w
        self._inv_sqrt_w = 1.0 / (self._sqrt_w)
        self._log_pdet = np.sum(np.log(self._w), axis=0).item()
        self._subspace_size = np.size(self._w)
        self._rank = self._subspace_size

    @property
    def eig_vects(self) -> NDArrayFloat:
        """Return the Eigen vectors."""
        return self._v

    @eig_vects.setter
    def eig_vects(self, value: NDArrayFloat) -> None:
        self._v = value
        self.n_pts = self._v.shape[0]

    def get_diagonal(self) -> NDArrayFloat:
        """
        Return the diagonal entries of the matrix (variances).
        """
        return np.sum(((self._v.T * np.sqrt(self._w)) ** 2), axis=0)

    def _todense(self) -> NDArrayFloat:
        return np.dot(self._v, np.multiply(self._w, self._v.T))

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return x @ (self._inv_sqrt_w * self._v.T).T

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        return x @ (self._sqrt_w * self._v.T)

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        return np.dot(
            self._v,
            np.multiply(self._w, np.dot(self._v.T, x.reshape(-1, 1))),
        )

    def _matmat(self, X: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        assert np.shape(X)[0] == self.shape[0]
        return np.dot(
            self._v,
            np.multiply(self._w, np.dot(self._v.T, X)),
        )

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        r"""
        Return $Q^{-1} b = ZD^{-1}Z^{T}b$.

        Parameters
        ----------
        b: NDArrayFloat
            Column vector with shape ($N_{\mathrm{s}}$, 1) or ensemble matrix with
            shape ($N_{\mathrm{s}}$, $N_e$).

        Returns
        -------
        NDAarrayFloat
            Column vector with shape ($N_{\mathrm{s}}$, 1) or ensemble matrix with
            shape ($N_{\mathrm{s}}$, $N_e$).
        """
        # np.dot(invZs.T, invZs)
        # Note: x must be a column vector of a matrix with size (Ns, Ne)
        ne = 1  # case of a column vector
        if b.ndim > 1:
            ne = b.shape[1]
        return np.dot(
            self._v,
            np.multiply(1.0 / self._w, np.dot(self._v.T, b.reshape(-1, ne))),
        )

    def get_sparse_LLT_factor(self) -> csc_array:
        """
        Return the sparse factor L of the LL^T factorization of the eigen matrix.

        Return
        ------
        L: csc_array
            L = U * V^{T/2}.
        """
        # 1) Convert U sqrt(V) to a sparse format
        sp_mat = sp.sparse.lil_array(self._v * np.sqrt(self._w).T)
        # 2) Resize -> we now have a square matrix and indices are preserved
        sp_mat.resize(self.shape)
        # 3) Convert to column format
        return sp_mat.tocsc()

    @property
    def precision(self) -> sp.sparse.sparray:
        return np.dot(
            self._v,
            np.multiply(1.0 / self._w, self._v.T),
        )


def get_eigen_factorization_gram_matrix_trick(anomalies: NDArrayFloat) -> NDArrayFloat:
    """
    _summary_

    Parameters
    ----------
    anomalies : NDArrayFloat
        _description_

    Returns
    -------
    NDArrayFloat
        _description_
    """
    ne, n_pts = np.shape(anomalies)
    if ne < n_pts:
        GramM = anomalies @ anomalies.T  # ne x ne matrix
    else:
        GramM = anomalies.T @ anomalies  # n_pts x n_pts matrix

    # Computed eigen decomposition
    # Could be randomized for large matrices
    eigvals, U = sp.linalg.eigh(GramM)

    # Sort
    idx = eigvals.argsort()[::-1]
    eigvals = eigvals[idx]
    U = U[:, idx]

    # Keep only positive values(truncate)
    tol = np.max(eigvals) * max(GramM.shape) * np.finfo(eigvals.dtype).eps
    mask = eigvals > tol

    eigvals = eigvals[mask]
    U = U[:, mask]

    # Rebuild the Eigen decomposition of anomalies
    s = np.sqrt(eigvals)

    if ne < n_pts:
        # U has shape (ne, n_pc)
        # anomalies has shape (ne, n_pts)
        V = (anomalies.T @ U) / s
    else:
        # U has shape (n_pts, p_pc)
        V = U
    # Vt has shape (n_pc, n_pts)
    return V, s


class CovViaEnsemble(CovarianceMatrix):
    r"""
    Represents a covariance matrix as an ensemble of realizations.

    For a given ensemble with shape (:math:`N_{s}`, :math:`N_{e}`), the number of
    points and the number of members in the ensemble respectively, the covariance
    matrix :math:`\mathbf{\Sigma_{ss}}` is approximated from the ensemble
    in the standard way of EnKF
    :cite:p:`evensenDataAssimilationEnsemble2007,aanonsenEnsembleKalmanFilter2009`:

    .. math::
        \mathbf{\Sigma_{ss}} = \frac{1}{N_{e} - 1} \sum_{j=1}^{N_{e}}\left(s_{j} -
        \overline{s}\right)\left(s_{j}
        - \overline{s^{l}} \right)^{T}

    Or by defining a matrix of anomalies
    :math:`\mathbf{A} = \mathbf{S} - \overline{\mathbf{S}}`
    with shape  (:math:`N_{s}`, :math:`N_{e}`):

    .. math::
        \mathbf{\Sigma_{ss}} = \frac{1}{N_{e} - 1} \mathbf{A}^{T}\mathbf{A}

    Note
    ----
    Practically, the dense covariance matrix is never built,
    only the anomalies matrix :math:`\mathbf{A}` is used. The product between the
    inverse of the covariance matrix and a vector
    :math:`\mathbf{x} = \mathbf{\Sigma_{ss}}^{-1}\mathbf{b}`
    is obtained solving the system :math:`\mathbf{A}^{T}\mathbf{Ax} = \mathbf{b}`,
    using gmres, where only anomalies matrix vector products are required.

    References
    ----------
    .. bibliography::
        :list: enumerated
        :filter: False

        evensenDataAssimilationEnsemble2007
        aanonsenEnsembleKalmanFilter2009

    """

    __slots__ = ["_ensemble", "_v", "_w", "_sqrt_w", "_inv_sqrt_w", "_inv_w"]

    def __init__(
        self,
        ensemble: NDArrayFloat,
    ) -> None:
        """
        Initiate the instance.

        Parameters
        ----------
        ensemble : NDArrayFloat
            Ensemble of realization with shape (:math:`N_{e}`, :math:`N_{s}`).
        """
        self.ensemble = ensemble
        super().__init__()

    @property
    def ensemble(self) -> NDArrayFloat:
        return self._ensemble

    @ensemble.setter
    def ensemble(self, value: NDArrayFloat) -> None:
        self._ensemble = value

        # build preconditioner using Eigen decomposition
        self._v, self._sqrt_w = get_eigen_factorization_gram_matrix_trick(
            self.anomalies / np.sqrt((self.n_ens - 1))
        )
        self._w = self._sqrt_w**2
        self._inv_w = 1.0 / self._w
        self._inv_sqrt_w = 1.0 / (self._sqrt_w)

        self.n_pts = np.shape(self.ensemble)[1]
        self._subspace_size = np.shape(self._v)[1]
        self._rank = 0
        self._log_pdet = 0.0

    @cached_property
    def anomalies(self) -> NDArrayFloat:
        """
        Return the matrix of anomalies.

        """
        return self.ensemble - np.mean(self.ensemble, axis=0, keepdims=True)

    @cached_property
    def n_ens(self) -> int:
        """Return the number of members in the ensemble."""
        return self.ensemble.shape[0]

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x (dot product)."""
        return np.linalg.multi_dot([self.anomalies.T, self.anomalies, x]) / (
            self.n_ens - 1
        )

    def _whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        Parameters
        ----------
        x : NDArrayFloat
            Data to whiten with shape (..., n), n being the number of points covered
            by the covariance matrix.

        Returns
        -------
        NDArrayFloat
            Transformed data with shape (..., n)
        """
        return x @ (self._v * self._inv_sqrt_w[None, :])

    def _colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        return x @ (self._sqrt_w[None, :].T * self._v.T)

    def _todense(self) -> NDArrayFloat:
        """
        Return a dense representation of the matrix.
        """
        return self.anomalies.T @ self.anomalies / (self.n_ens - 1)

    def solve(
        self, b: NDArrayFloat, rtol: float = 1e-12, maxiter: int = 1000
    ) -> NDArrayFloat:
        """
        Solve A^{T}Ax = b, with A, the anomalies matrix instance.

        Note that the dense covariance matrix is never built.

        TODO: explains that it relies on the eigen decomposition.
        """
        return np.linalg.multi_dot(
            [
                (self._v * self._inv_w[None, :]),
                self._v.T,
                b,
            ]
        )

    def get_diagonal(self) -> NDArrayFloat:
        """Return the diagonal entries of the matrix (variances)."""
        return np.sum((self.anomalies**2), axis=0) / (self.n_ens - 1.0)

    @property
    def precision(self) -> NDArrayFloat:
        return (self._v * self._inv_w[None, :]) @ self._v.T


def _generate_dense_matrix_from_kernel(
    pts: NDArrayFloat, kernel: Callable, len_scale: NDArrayFloat, nugget: float = 0.0
) -> NDArrayFloat:
    """
    Generate a dense matrix.

    Compute O(dim^2) interactions.

    Parameters
    ----------
    pts : NDArrayFloat
        DESCRIPTION.
    kernel : Callable
        DESCRIPTION.
    len_scale: NDArrayFloat
        DESCRIPTION.
    nugget: float
        It acts as a regularization. The default is 0.0 (no nugget effect).

    Returns
    -------
    NDArrayFloat
        The dense matrix.
    """
    # Scale the points coordinates
    scaled_pts = np.array(pts, copy=True)
    for dim in range(scaled_pts.shape[1]):
        scaled_pts[:, dim] /= len_scale[dim]
    return kernel(sp.spatial.distance_matrix(scaled_pts, scaled_pts))


class CovKernelAsLinop(abc.ABC, LinearOperator):
    """Abstract class providing linear operator capability from a kernel definition."""

    __slots__: List[str] = [
        "_kernel",
        "_pts",
        "_nugget",
        "_preconditioner",
        "count",
        "solvematvecs",
    ]

    def __init__(
        self,
        pts: NDArrayFloat,
        kernel: Callable,
        len_scale: NDArrayFloat,
        nugget: float = 0.0,
        k: int = 100,
        is_use_preconditioner: bool = False,
    ) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        pts : NDArrayFloat
            _description_
        kernel : Callable
            _description_
        nugget : float, optional
            _description_, by default 0.0
        """
        self._kernel: Callable = kernel
        self._pts: NDArrayFloat = pts
        self._nugget: float = nugget
        self._len_scale: NDArrayFloat = len_scale
        # counters
        self.count: int = 0
        self.solvmatvecs: int = 0

        super().__init__(shape=(self.n_pts, self.n_pts), dtype="d")

        if is_use_preconditioner:
            self._preconditioner: Optional[csr_array] = _build_kernel_preconditioner(
                pts, kernel, k=k
            )
        else:
            self._preconditioner = None

    @property
    def n_pts(self) -> int:
        """Return the number of points covered."""
        return np.shape(self._pts)[0]

    def todense(self) -> NDArrayFloat:
        """
        Explicit dense representation of the covariance matrix.
        """
        return _generate_dense_matrix_from_kernel(
            self._pts, self._kernel, self._len_scale, self._nugget
        )

    def get_diagonal(self) -> NDArrayFloat:
        """Return the diagonal entries of the matrix (variances)."""
        return self._kernel(np.zeros(len(self._pts)))

    def reset_comptors(self) -> None:
        """Set the comptors to zero."""
        self.count = 0
        self.solvmatvecs = 0

    def itercount(self) -> int:
        """Return the number of counts."""
        return self.count

    def solve(
        self, b: NDArrayFloat, rtol: float = 1e-12, maxiter: int = 1000
    ) -> NDArrayFloat:
        """
        Solve Ax = b, with A, the current covariance matrix instance.

        Notes
        -----
        It relies on GMRES and matrix-vector multiplications (column by column).

        Parameters
        ----------
        b : NDArrayFloat
            Array with shape (n,), (n, 1) or (n, ne), ne being the number of
            vectors (columns). `x` will have the same shape as b.
        rtol : float, optional
            Relative tolerance for the convergence of GMRES, by default 1e-12
        maxiter : int, optional
            Maximum number of GMRES iterations, by default 1000

        Returns
        -------
        NDArrayFloat
            The x array with same shape as b.

        """

        residual = CallBack()
        # make sure to have 2d array
        _b = np.reshape(b, (self.n_pts, -1))

        nv = np.shape(_b)[1]
        x = np.zeros_like(_b)

        # does not work for matrices ?
        for i in range(nv):
            x[:, i], info = gmres(
                self,
                b=_b[:, i],
                rtol=rtol,
                maxiter=maxiter,
                callback=residual,
                callback_type="legacy",
                M=self._preconditioner,
                atol=0.0,
            )
            self.solvmatvecs += residual.itercount

        # Get the original shape back
        return x.reshape(np.shape(b))


def _build_kernel_preconditioner(
    pts: NDArrayFloat, kernel: Callable, k: int = 100
) -> csr_array:
    """
    Implementation of the preconditioner based on changing basis.

    Parameters
    ----------
    pts : NDArrayFloat
        The points (n, m) with n the number of data points and m the dimension of
        coordinates.
    kernel: Callable
        The covariance kernel defined as a callable instance.
    k : int, optional
        Number of local centers in the preconditioner. Controls the sparity of
        the preconditioner. By default 100.

    Returns
    -------
    csr_array
        _description_

    Raises
    ------
    ValueError
        _description_

    Notes:
    ------
    Implementation of the preconditioner based on local centers.
    The parameter k controls the sparsity and the effectiveness of the preconditioner.
    Larger k is more expensive but results in fewer iterations.
    For large ill-conditioned systems, it was best to use a nugget effect to make the
    problem better conditioned.
    To Do: implementation based on local centers and additional points. Will remove the
    hack of using nugget effect.

    """
    nb_pts: int = pts.shape[0]
    if nb_pts <= 0:
        raise ValueError("The number of points cannot be null !")
    if nb_pts < k:
        raise ValueError(
            f"k ({k}) must be lower or equal to the number of points ({nb_pts})!"
        )

    # Build the tree
    start: float = time()
    tree: sp.spatial.cKDTree = sp.spatial.cKDTree(pts, leafsize=32)
    end: float = time()

    logging.log(logging.INFO, f"Tree building time = {end - start}")

    # Find the nearest neighbors of all the points
    start = time()
    _dist, ind = tree.query(pts, k=k)
    end = time()

    logging.log(logging.INFO, f"Nearest neighbor computation time = {end - start}")

    Q = np.zeros((k, k), dtype="d")
    y = np.zeros((k, 1), dtype="d")

    row = np.tile(np.arange(nb_pts), (k, 1)).transpose()
    col = np.copy(ind)
    nu = np.zeros((nb_pts, k), dtype="d")

    y[0] = 1.0
    start = time()

    # TODO: This is very inefficient and must be re-written
    for i in np.arange(nb_pts):
        Q = kernel(cdist(pts[ind[i, :], :], pts[ind[i, :], :]))
        nui = sp.linalg.solve(Q, y)
        nu[i] = nui.ravel()

    end = time()

    logging.log(logging.INFO, "Elapsed time = %g" % (end - start))

    ij = np.zeros((nb_pts * k, 2), dtype="i")
    ij[:, 0] = np.copy(np.reshape(row, nb_pts * k, order="F").transpose())
    ij[:, 1] = np.copy(np.reshape(col, nb_pts * k, order="F").transpose())

    data = np.copy(np.reshape(nu, nb_pts * k, order="F").transpose())
    return csr_array((data, ij.transpose()), shape=(nb_pts, nb_pts), dtype="d")


class CovKernelAsLinopViaFFT(CovKernelAsLinop):
    """
    Represents a fast fourier transform covariance matrix.

    FFT based operations if kernel is stationary or translation invariant and points
    are on a regular grid.

    Notes
    -----
    This FFT-based code is a reimplementation of Arvind Saibaba's work
    (https://github.com/arvindks/kle).

    TODO: add references.
    """

    __slots__: List[str] = ["_first_row"]

    def __init__(
        self,
        kernel,
        mesh_dim: Union[float, NDArrayFloat, Sequence[float]],
        domain_shape: Union[int, NDArrayInt, Sequence[int]],
        len_scale: NDArrayFloat,
        nugget: float = 0.0,
        k: int = 100,
        is_use_preconditioner: bool = False,
    ) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        kernel : _type_
            _description_
        mesh_dim : Union[NDArrayInt, Tuple[float, float]]
            _description_
        domain_shape : Union[NDArrayInt, Tuple[int, int]]
            _description_
        len_scale : NDArrayFloat
            _description_
        nugget : float, optional
            _description_, by default 0.0
        k : int, optional
            Number of local centers in the preconditioner. Controls the sparity of
            the preconditioner. It should be inferior to the number of points.
            By default 100.
        is_use_preconditioner: bool
            Whether to build the preconditioner at instance creation and use it to
            solve Ax = b systems. The default is False.
        """
        self.param_shape: NDArrayInt = np.array(domain_shape, dtype=np.int8)
        # Coordinates of the points in the grid with shape (Npts, Ndim)
        pts = get_pts_coords_regular_grid(mesh_dim, self.param_shape)

        self._first_row = create_toepliz_first_row(pts, kernel, len_scale)
        super().__init__(
            pts,
            kernel,
            len_scale=len_scale,
            nugget=nugget,
            k=k,
            is_use_preconditioner=is_use_preconditioner,
        )

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return the covariance matrix times the vector x."""
        return toeplitz_product(x, self._first_row, self.param_shape) * (
            1 + self._nugget
        )


def get_linop_eigen_factorization(
    linop: LinearOperator,
    size: int,
    n_pc: int,
    random_state: Optional[Union[int, RandomState, Generator]] = None,
) -> Tuple[NDArrayFloat, NDArrayFloat]:
    """
    Compute Eigenmodes of the covariance.

    Parameters
    ----------
    linop: LinearOperator
        The covariance matrix as a linear operator to factorize.
    size: int
        Input vector size for the linear operator.
    n_pc : int
        Number of principal component in the matrix.
    random_state: Optional[Union[int, np.random.Generator, np.random.RandomState]]
        Pseudorandom number generator state used to generate resamples.
        If `random_state` is ``None`` (or `np.random`), the
        `numpy.random.RandomState` singleton is used.
        If `random_state` is an int, a new ``RandomState`` instance is used,
        seeded with `random_state`.
        If `random_state` is already a ``Generator`` or ``RandomState``
        instance then that instance is used.

    Raises
    ------
    NotImplementedError
        If a method difference from arpack is used for decomposition.

    Returns
    -------
    Tuple[NDArrayFloat, NDArrayFloat]
        Eigen values and eigen vectors.
    """
    logging.info("Eigen factorization of linear operator")

    # twopass = False if not 'twopass' in self.params else self.params['twopass']
    start = time()

    # Random state for v0 vector used by eigsh and svds
    if random_state is not None:
        random_state = check_random_state(random_state)
        v0 = random_state.uniform(size=(size,))
    else:
        v0 = None

    eig_vals, eig_vects = sp.sparse.linalg.eigsh(linop, k=n_pc, v0=v0)
    eig_vals = eig_vals[::-1]
    eig_vals = eig_vals.reshape(-1, 1)  # make a column vector
    eig_vects = eig_vects[:, ::-1]

    logging.info(
        "- time for eigenfactorization with k = %d is %g sec"
        % (n_pc, round(time() - start))
    )

    if (eig_vals > 0).sum() < n_pc:
        n_pc = (eig_vals > 0).sum()
        eig_vals = eig_vals[:n_pc, :]
        eig_vects = eig_vects[:, :n_pc]
        logging.warning("Warning: n_pc changed to %d for positive eigenvalues" % (n_pc))

    logging.info(
        f"- 1st eigv : {eig_vals[0]}, {n_pc}-th eigv : {eig_vals[-1]}, "
        f"ratio: {eig_vals[-1] / eig_vals[0]}"
    )
    return eig_vals, eig_vects


def eigen_factorize_cov_mat(
    cov_mat: Union[CovarianceMatrix, CovKernelAsLinop],
    n_pc: int,
    random_state: Optional[Union[int, RandomState, Generator]] = None,
) -> CovViaEigenFactorization:
    """
    Return an eigen factorized covariance matrix from the input covariance matrix.

    Parameters
    ----------
    cov_mat : Union[CovarianceMatrix, CovKernelAsLinop]
        The covariance matrix instance to decompose.
    n_pc : int
        Number of principal component in the matrix.
    random_state: Optional[Union[int, np.random.Generator, np.random.RandomState]]
        Pseudorandom number generator state used to generate resamples.
        If `random_state` is ``None`` (or `np.random`), the
        `numpy.random.RandomState` singleton is used.
        If `random_state` is an int, a new ``RandomState`` instance is used,
        seeded with `random_state`.
        If `random_state` is already a ``Generator`` or ``RandomState``
        instance then that instance is used.

    Returns
    -------
    CovViaEigenFactorization
        Decomposed matrix instance.
    """
    if isinstance(cov_mat, CovViaEigenFactorization):
        return cov_mat
    return CovViaEigenFactorization(
        get_linop_eigen_factorization(cov_mat, cov_mat.shape[0], n_pc, random_state)
    )


def get_explained_var(
    eigval: NDArrayFloat,
    cov_mat: Optional[CovarianceMatrix] = None,
    trace_cov_mat: Optional[float] = None,
) -> NDArrayFloat:
    """Return the variance explained by each eigen value."""
    if trace_cov_mat is not None:
        return eigval / trace_cov_mat
    if cov_mat is not None:
        return eigval / cov_mat.get_trace()
    else:
        raise ValueError("You must provide a Covariance matrix instance or the trace !")



from typing import no_type_check

import scipy as sp


@no_type_check
def _get_L_D_P(A: sp.sparse.sparray):
    """
    Return L, D and P from the factorization L @ D @ L' = P @ A @ P' using sksparse.

    Note that sksparse uses SuiteSparse which is LGPL licence.
    """
    import sksparse.cholmod as cholmod

    # Need to take the API change into account
    try:
        # sksparse 4.x
        L, D, P = cholmod.ldl(A, order="amd")
    except AttributeError:
        # sksparse 5.x
        f = cholmod.cholesky(A)
        (L, D), P = f.L_D(), f.P()
    return L, D, P



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Antoine COLLET

"""Provide utils to work with list."""

import numpy as np
import numpy.typing as npt

NDArrayFloat = npt.NDArray[np.float64]
NDArrayInt = npt.NDArray[np.int64]
NDArrayBool = npt.NDArray[np.bool_]
ArrayLike = npt.ArrayLike



"""Some tests to refactor."""

import re

import covmats
import numpy as np
import pytest
import scipy as sp
from covmats._covariances import (
    CallBack,
    _build_kernel_preconditioner,
    get_pts_coords_regular_grid,
)
from covmats._sparse_helpers import get_SPD_sparse_n11_example
from covmats._types import NDArrayFloat
from covmats.data import load_precision_example_4225x

from .sparse_helpers import _get_L_D_P  # ty:ignore[unresolved-import]

v3 = np.array([3.0, 1.0, 8.0])
V34 = np.vstack([v3] * 4).T
assert np.shape(V34) == (3, 4)

v5 = np.array([3.0, 0.0, 8.0, 9.76, -1.87])
V54 = np.vstack([v5] * 4).T
assert np.shape(V54) == (5, 4)


def test_CallBack():
    c = CallBack()
    assert c.itercount == 0
    for i in range(3):
        c(np.ones(10))
    assert c.itercount == 3
    c.clear()
    assert c.itercount == 0


def test_removed_from_cholesky() -> None:

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "`from_cholesky` is not available, please instantiate"
            " with `CovViaCholesky(...)` directly!"
        ),
    ):
        covmats.CovarianceMatrix.from_cholesky()


def test_removed_from_diagonal() -> None:

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "`from_diagonal` is not available, please instantiate"
            " with `CovViaDiagonal(...)` directly!"
        ),
    ):
        covmats.CovarianceMatrix.from_diagonal()


def test_removed_from_eigendecomposition() -> None:

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "`from_eigendecomposition` is not available, please instantiate"
            " with `CovViaEigenFactorization(...)` directly!"
        ),
    ):
        covmats.CovarianceMatrix.from_eigendecomposition()


def test_removed_from_precision() -> None:

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "`from_precision` is not available, please instantiate"
            " with `CovViaPrecisionCholesky(...)` directly!"
        ),
    ):
        covmats.CovarianceMatrix.from_precision()


def test_validate_dense_matrix() -> None:
    with pytest.raises(
        ValueError,
        match=(
            "The input `my_arg` must be a square, "
            "two-dimensional array of real numbers."
        ),
    ):
        covmats.CovarianceMatrix._validate_dense_matrix(np.ones((10, 9)), "my_arg")

    # no issue
    covmats.CovarianceMatrix._validate_dense_matrix(np.ones((10, 10)), "my_arg")


def test_validate_sparse_matrix():
    with pytest.raises(
        ValueError,
        match=(
            "The input `my_arg` must be a square, "
            "two-dimensional array of real numbers."
        ),
    ):
        covmats.CovarianceMatrix._validate_sparse_matrix(
            sp.sparse.csc_array((10, 9)), "my_arg"
        )

    # No issue
    covmats.CovarianceMatrix._validate_sparse_matrix(
        sp.sparse.csc_array((10, 10)), "my_arg"
    )


def test_validate_vector():
    with pytest.raises(
        ValueError,
        match=("The input `my_arg` must be a one-dimensional array of real numbers."),
    ):
        covmats.CovarianceMatrix._validate_vector(np.ones((2, 2)), "my_arg")

    with pytest.raises(
        ValueError,
        match=("The input `my_arg` must be a one-dimensional array of real numbers."),
    ):
        covmats.CovarianceMatrix._validate_vector(np.ones((2, 1)), "my_arg")

    # No issue
    covmats.CovarianceMatrix._validate_vector(np.ones(((2,))), "my_arg")


def test_validate_dense_lower_triangle():
    with pytest.raises(
        ValueError,
        match=("The input `D` must be a lower-triangular matrix."),
    ):
        covmats.CovarianceMatrix._validate_dense_lower_triangle(np.ones((2, 2)), "D")

    # No issue
    covmats.CovarianceMatrix._validate_dense_lower_triangle(
        np.array([[1.0, 0.0], [1.0, 1.0]]), "D"
    )


def test_CovViaDiagonal_singular_or_indefinite() -> None:

    with pytest.raises(
        ValueError,
        match=(
            "2 null values have been detected in `D` which "
            "means the matrix is singular and non invertible."
        ),
    ):
        covmats.CovViaDiagonal(np.array([1.0, 0.0, 0.0, 8.789]))

    with pytest.raises(
        ValueError,
        match=(
            "1 negative values have been detected in `D` which "
            "means the matrix is indefinite and non invertible."
        ),
    ):
        covmats.CovViaDiagonal(np.array([1.0, 0.001, -1.9, 8.789]))


def test_CovViaDiagonal_stats1() -> None:

    d = [1, 2, 3]
    A33 = np.diag(d)  # a diagonal covariance matrix
    x = [4, -2, 5]  # a point of interest
    ref_dist = sp.stats.multivariate_normal(mean=[0, 0, 0], cov=A33)

    # Make a diagonal covariance matrix
    cov_diag33 = covmats.CovViaDiagonal(d)
    assert cov_diag33.rank == 3
    np.testing.assert_allclose(cov_diag33.log_pdet, 1.791759, rtol=1e-6)
    assert cov_diag33.get_trace() == 6

    cov_dist = sp.stats.multivariate_normal(mean=[0, 0, 0], cov=cov_diag33)
    np.testing.assert_allclose(ref_dist.pdf(x), cov_dist.pdf(x))

    # Test to dense and covariance
    np.testing.assert_allclose(cov_diag33.todense(), A33)
    np.testing.assert_allclose(cov_diag33.covariance, A33)


def test_CovViaDiagonal_stats2() -> None:

    rng = np.random.default_rng(2026)
    n = 5
    A = np.diag(rng.random(n))
    x = rng.random(size=n)

    d = np.diag(A)
    cov = covmats.CovViaDiagonal(d)

    res = cov.whiten(x)
    ref = np.diag(d**-0.5) @ x
    assert np.allclose(res, ref)

    res = cov.log_pdet
    ref = np.linalg.slogdet(A)[-1]
    assert np.allclose(res, ref)


def test_CovViaDiagonal_aslinop() -> None:

    d = [1, 2, 3]
    # Make a diagonal covariance matrix
    cov_diag33 = covmats.CovViaDiagonal(d)

    # matvec and rmatvec
    np.testing.assert_allclose(cov_diag33 @ v3, np.array([3.0, 2.0, 24.0]))
    np.testing.assert_allclose(cov_diag33 @ v3, cov_diag33.T @ v3)

    # matmat
    np.testing.assert_allclose(
        cov_diag33 @ V34, np.vstack([np.array([3.0, 2.0, 24.0])] * 4).T
    )

    # rmatmat
    np.testing.assert_allclose(cov_diag33 @ V34, cov_diag33.T @ V34)

    # solve
    np.testing.assert_allclose(cov_diag33.solve(cov_diag33 @ v3), v3)
    np.testing.assert_allclose(cov_diag33.solve(cov_diag33 @ V34), V34)

    # Get precision matrix (inverse)
    np.testing.assert_allclose(
        np.linalg.inv(cov_diag33.todense()), cov_diag33.precision
    )


def test_CovViaDiagonal_mvnormal() -> None:

    covd = covmats.CovViaDiagonal(np.array([5.0, 10.0, 15.0]))
    rng_seed = 42
    covd.sample_mvnormal(shape=[2], random_state=rng_seed)
    x = covd.sample_mvnormal(shape=[2, 4], random_state=rng_seed)
    np.testing.assert_allclose(
        x,
        np.array(
            [
                [
                    [1.11068661, -0.43723011, 2.50848692],
                    [3.40559829, -0.74045799, -0.90680853],
                    [3.53122721, 2.4268417, -1.81826648],
                    [1.21320114, -1.46545542, -1.80376358],
                ],
                [
                    [0.54104409, -6.05032338, -6.68057804],
                    [-1.25731314, -3.20285323, 1.21707469],
                    [-2.03040356, -4.46609644, 5.67643327],
                    [-0.50485116, 0.21354293, -5.518026],
                ],
            ]
        ),
    )
    assert x.shape == (2, 4, 3)


def test_CovViaCholesky_log_stats() -> None:

    rng = np.random.default_rng(2026)
    n = 5
    A = rng.random(size=(n, n))
    A = A @ A.T  # make the covariance symmetric positive definite
    x = rng.random(size=n)

    L = np.linalg.cholesky(A)
    cov = covmats.CovViaCholesky(L)

    res = cov.whiten(x)
    ref = sp.linalg.solve_triangular(L, x, lower=True)
    assert np.allclose(res, ref)

    res = cov.log_pdet
    ref = np.linalg.slogdet(A)[-1]
    assert np.allclose(res, ref)


def test_CovViaCholesky_aslinop() -> None:

    rng = np.random.default_rng(2026)
    n = 5
    A = rng.random(size=(n, n))
    A = A @ A.T  # make the covariance symmetric positive definite

    L = np.linalg.cholesky(A)
    cov = covmats.CovViaCholesky(L)

    # Check to dense
    np.testing.assert_allclose(cov.todense(), A)

    expected = np.array([20.932923, 33.649514, 42.873409, 25.55103, 22.343875])

    # matvec and rmatvec
    np.testing.assert_allclose(cov @ v5, expected, rtol=1e-6)
    np.testing.assert_allclose(cov @ v5, cov.T @ v5)

    # matmat
    np.testing.assert_allclose(cov @ V54, np.vstack([expected] * 4).T, rtol=1e-6)

    # rmatmat
    np.testing.assert_allclose(cov @ V54, cov.T @ V54)

    # solve
    np.testing.assert_allclose(cov.solve(cov @ v5), v5, atol=1e-10)
    np.testing.assert_allclose(cov.solve(cov @ V54), V54, atol=1e-10)

    # Get precision matrix (inverse)
    np.testing.assert_allclose(np.linalg.inv(A), cov.precision)

    # Test to dense with original covariance stored
    cov2 = covmats.CovViaCholesky(L, covariance=A)
    np.testing.assert_allclose(cov2.todense(), A)

    # Test  diagonal
    np.testing.assert_allclose(cov2.get_diagonal(), np.diagonal(A))


def test_CovViaCholesky_mvnormal() -> None:

    covd = covmats.CovViaDiagonal(np.array([5.0, 10.0, 15.0]))
    cov_cho = covmats.CovViaCholesky(sp.linalg.cholesky(covd.todense()))

    rng_seed = 42
    covd.sample_mvnormal(shape=[2], random_state=rng_seed)
    x = cov_cho.sample_mvnormal(shape=[2, 4], random_state=rng_seed)
    np.testing.assert_allclose(
        x,
        np.array(
            [
                [
                    [1.11068661, -0.43723011, 2.50848692],
                    [3.40559829, -0.74045799, -0.90680853],
                    [3.53122721, 2.4268417, -1.81826648],
                    [1.21320114, -1.46545542, -1.80376358],
                ],
                [
                    [0.54104409, -6.05032338, -6.68057804],
                    [-1.25731314, -3.20285323, 1.21707469],
                    [-2.03040356, -4.46609644, 5.67643327],
                    [-0.50485116, 0.21354293, -5.518026],
                ],
            ]
        ),
    )
    assert x.shape == (2, 4, 3)


@pytest.mark.parametrize("is_embbed_covariance", (True, False))
def test_CovViaSparseCholesky(is_embbed_covariance: bool) -> None:

    N: int = 11
    # precision matrix (sparse)
    A = get_SPD_sparse_n11_example(seed=2026)
    # dense covariance matrix
    Q = np.linalg.inv(A.toarray())
    cov = covmats.CovViaSparseCholesky(
        covmats.SparseCholeskyFactor(*_get_L_D_P(A)),
        sparse_covariance=A if is_embbed_covariance else None,
    )

    # test access scf
    cov.scf

    # Test to dense
    np.testing.assert_allclose(cov.todense(), A.toarray())
    # Test shape
    assert cov.shape == (N, N)

    # Diagonal
    np.testing.assert_allclose(cov.get_diagonal(), A.diagonal())

    # Dense
    np.testing.assert_allclose(cov.precision, Q)
    np.testing.assert_allclose(cov.todense(), A.toarray())

    # Test solve
    expected_x = np.arange(N, dtype=np.float64)
    b = A @ expected_x
    np.allclose(cov @ b, expected_x)
    np.allclose(cov.solve(b), expected_x)

    # Test colorize and whiten
    np.testing.assert_allclose(
        cov.whiten(cov.colorize(np.eye(N))), np.eye(N), atol=1e-15
    )

    # Test colorize with an ensemble
    colored_samples = cov.sample_mvnormal(
        shape=(500_000,), random_state=np.random.default_rng(2027)
    )
    np.testing.assert_allclose(
        covmats.CovViaEnsemble(colored_samples).todense(), A.toarray(), atol=2e-1
    )

    # Logdet
    np.testing.assert_allclose(np.linalg.slogdet(A.toarray())[1], cov.log_pdet)


@pytest.mark.parametrize("is_embbed_precision", (True, False))
def test_CovViaPrecisionCholesky(is_embbed_precision: bool) -> None:

    rng = np.random.default_rng()
    n = 5
    P = rng.random(size=(n, n))
    Q = P @ P.T  # a precision matrix must be positive definite
    x = rng.random(size=n)

    A = np.linalg.inv(Q)

    cov = covmats.CovViaPrecisionCholesky(
        np.linalg.cholesky(Q), precision=Q if is_embbed_precision else None
    )

    res = cov.whiten(x)
    ref = x @ np.linalg.cholesky(Q)
    assert np.allclose(res, ref)

    res = cov.log_pdet
    ref = -np.linalg.slogdet(Q)[-1]
    assert np.allclose(res, ref)

    # Test to dense
    np.testing.assert_allclose(cov.todense(), A)
    # Test shape
    assert cov.shape == (n, n)

    # Diagonal
    np.testing.assert_allclose(cov.get_diagonal(), A.diagonal())

    # Dense
    np.testing.assert_allclose(cov.precision, Q)
    np.testing.assert_allclose(cov.todense(), A)

    # Test matvec and solve
    expected_x = np.arange(n, dtype=np.float64)
    b = A @ expected_x
    np.allclose(cov @ b, expected_x)
    np.allclose(cov.solve(b), expected_x)

    # Test colorize and whiten
    np.testing.assert_allclose(
        cov.whiten(cov.colorize(np.eye(n))), np.eye(n), atol=1e-10, rtol=1e-10
    )

    # Test colorize with an ensemble
    colored_samples = cov.sample_mvnormal(
        shape=(500_000,), random_state=np.random.default_rng(2027)
    )
    np.testing.assert_allclose(
        covmats.CovViaEnsemble(colored_samples).todense(), A, rtol=2e-1, atol=5e-2
    )

    np.testing.assert_allclose(np.linalg.slogdet(A)[1], cov.log_pdet)


def test_CovViaPrecisionCholesky_whiten() -> None:

    rng = np.random.default_rng(2026)
    n = 3
    A = rng.random(size=(n, n))
    cov_array = A @ A.T  # make matrix symmetric positive definite
    precision = np.linalg.inv(cov_array)
    cov_object = covmats.CovViaPrecisionCholesky(
        sp.linalg.cholesky(precision, lower=True)
    )
    x = rng.multivariate_normal(np.zeros(n), cov_array, size=(10000))
    x_ = cov_object.whiten(x)
    # near-identity covariance is expected
    np.testing.assert_allclose(np.cov(x_, rowvar=False), np.eye(3), atol=0.01)


def test_ensemble_covariance_matrix() -> None:
    """Test the inversion."""

    cov = covmats.CovViaEnsemble(np.random.default_rng(2023).random((200, 77)))
    x = np.random.default_rng(2023).random(77)

    np.testing.assert_allclose(cov.solve(x), np.linalg.inv(cov.todense()).dot(x))
    np.testing.assert_allclose(np.trace(cov.todense()), cov.get_trace(), rtol=1e-12)


@pytest.mark.parametrize("is_embbed_precision", (True, False))
def test_CovViaSparsePrecisionCholesky(is_embbed_precision: bool) -> None:

    N: int = 11
    # precision matrix (sparse)
    Q = get_SPD_sparse_n11_example(seed=2026)
    # dense covariance matrix
    A = np.linalg.inv(Q.toarray())
    cov = covmats.CovViaSparsePrecisionCholesky(
        covmats.SparseCholeskyFactor(*_get_L_D_P(Q)),
        sparse_precision=Q if is_embbed_precision else None,
    )

    # test access scf
    cov.scf

    # Test to dense
    np.testing.assert_allclose(cov.todense(), A)
    # Test shape
    assert cov.shape == (N, N)

    # Diagonal
    np.testing.assert_allclose(cov.get_diagonal(), A.diagonal())

    # Dense
    np.testing.assert_allclose(cov.precision.todense(), Q.toarray())
    np.testing.assert_allclose(cov.todense(), A)

    # Test solve
    expected_x = np.arange(N, dtype=np.float64)
    b = A @ expected_x
    np.allclose(cov @ expected_x, A @ expected_x)
    np.allclose(cov.solve(b), expected_x)

    # Test colorize and whiten
    np.testing.assert_allclose(
        cov.whiten(cov.colorize(np.eye(N))), np.eye(N), atol=1e-15
    )

    # Test colorize with an ensemble
    colored_samples = cov.sample_mvnormal(
        shape=(500_000,), random_state=np.random.default_rng(2027)
    )
    np.testing.assert_allclose(
        covmats.CovViaEnsemble(colored_samples).todense(), A, atol=1e-2
    )

    np.testing.assert_allclose(np.linalg.slogdet(A)[1], cov.log_pdet)


def test_build_kernel_preconditioner() -> None:

    _number_grid_cells = 225
    prior_std = 2.0

    # Exponential covariance model
    def exponential_kernel(r: float) -> NDArrayFloat:
        return (prior_std**2) * np.exp(-r)

    param_shape = np.array(
        [np.sqrt(_number_grid_cells), np.sqrt(_number_grid_cells)], dtype=np.int8
    )
    # _params = {"R": 1.0e-4, "kappa": 100}
    dx = 1.0 / 50.0
    dy = 1.0 / 50.0
    mesh_dim = (dx, dy)

    pts = get_pts_coords_regular_grid(mesh_dim, param_shape)

    # No issue
    _build_kernel_preconditioner(pts, exponential_kernel, k=100)

    # Issue
    with pytest.raises(ValueError, match="The number of points cannot be null !"):
        _build_kernel_preconditioner(np.array([]), exponential_kernel, k=100)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "k (1000) must be lower or equal to the number of points (225)!"
        ),
    ):
        _build_kernel_preconditioner(np.array(pts), exponential_kernel, k=1000)


def test_fft_covariance_matrix() -> None:
    _number_grid_cells = 225
    prior_std = 2.0

    # Exponential covariance model
    def exponential_kernel(r: float) -> NDArrayFloat:
        return (prior_std**2) * np.exp(-r)

    param_shape = np.array(
        [np.sqrt(_number_grid_cells), np.sqrt(_number_grid_cells)], dtype=np.int8
    )
    # _params = {"R": 1.0e-4, "kappa": 100}
    dx = 1.0 / 50.0
    dy = 1.0 / 50.0
    len_scale = np.array([1, 1])
    mesh_dim = (dx, dy)

    cov = covmats.CovKernelAsLinopViaFFT(
        exponential_kernel,
        mesh_dim=mesh_dim,
        domain_shape=param_shape,
        len_scale=len_scale,
        nugget=1e-4,
        is_use_preconditioner=True,
    )

    # tests
    assert cov.n_pts == 225
    np.testing.assert_allclose(cov.get_diagonal(), np.ones(_number_grid_cells) * 4.0)
    assert np.sum(cov.get_diagonal()) == 900

    # reinitiate comptors
    cov.reset_comptors()
    assert cov.itercount() == 0

    # Dense version
    np.testing.assert_allclose(cov @ np.eye(cov.n_pts), cov.todense(), rtol=1e-3)

    # Test to matvec and solve
    np.testing.assert_allclose(cov.solve(cov @ np.ones(225)), np.ones(225))


@pytest.mark.parametrize("is_use_preconditioner,", ((True), (False)))
def test_eigen_decompose_and_associated_functions(is_use_preconditioner: bool) -> None:
    _number_grid_cells = 225
    prior_std = 2.0

    # Exponential covariance model
    def exponential_kernel(r: float) -> NDArrayFloat:
        return (prior_std**2) * np.exp(-r)

    param_shape = np.array(
        [np.sqrt(_number_grid_cells), np.sqrt(_number_grid_cells)], dtype=np.int8
    )
    # _params = {"R": 1.0e-4, "kappa": 100}
    dx = 1.0 / 50.0
    dy = 1.0 / 50.0
    len_scale = np.array([1, 1])
    mesh_dim = (dx, dy)

    cov_mat_fft = covmats.CovKernelAsLinopViaFFT(
        exponential_kernel,
        mesh_dim=mesh_dim,
        domain_shape=param_shape,
        len_scale=len_scale,
        nugget=1e-4,
        is_use_preconditioner=is_use_preconditioner,
    )

    eig_mat = covmats.eigen_factorize_cov_mat(cov_mat_fft, n_pc=100, random_state=25652)
    assert eig_mat.n_pc == 100
    # should return the matrix as is
    eig_mat = covmats.eigen_factorize_cov_mat(eig_mat, 50)
    assert eig_mat.n_pc == 100  # and not 50 !

    # no random state for the test
    _ = covmats.get_linop_eigen_factorization(eig_mat, 50, n_pc=12)

    # This is determined form the eigen vectors
    assert eig_mat.n_pts == 225

    np.testing.assert_allclose(
        eig_mat.get_diagonal(), cov_mat_fft.get_diagonal(), rtol=0.05
    )
    # The trace should be around 900 (225 * 2.0 ** 2)
    np.testing.assert_allclose(eig_mat.get_trace(), 900, rtol=0.05)

    samples = eig_mat.sample_mvnormal(shape=(20,), random_state=2026)
    assert samples.shape == (20, 225)
    samples = eig_mat.sample_mvnormal(
        shape=(
            10,
            7,
            99,
        ),
        random_state=2026,
    )
    assert samples.shape == (10, 7, 99, 225)
    assert eig_mat.todense().shape == (225, 225)

    eig_mat.precision
    # np.testing.assert_allclose(
    #     eig_mat.precision, np.linalg.inv(cov_mat_fft.todense()), rtol=0.1
    # )

    L = eig_mat.get_sparse_LLT_factor()
    np.testing.assert_allclose((L @ L.T).toarray(), eig_mat.todense())

    # Test to dense
    np.testing.assert_allclose(cov_mat_fft.todense(), eig_mat.todense(), rtol=1e-2)
    # Test trace and diagonal
    _trace = eig_mat.get_trace()
    np.testing.assert_allclose(_trace, np.sum(cov_mat_fft.get_diagonal()), rtol=1e-2)

    # both covariance matrice instance and trace
    covmats.get_explained_var(eig_mat.eig_vals, eig_mat, _trace)
    # no trace
    covmats.get_explained_var(eig_mat.eig_vals, eig_mat)
    # no matrix
    covmats.get_explained_var(eig_mat.eig_vals, trace_cov_mat=_trace)
    # none
    with pytest.raises(
        ValueError, match="You must provide a Covariance matrix instance or the trace !"
    ):
        covmats.get_explained_var(eig_mat.eig_vals)

    # Test linop capability
    test_v = np.ones(225)
    eig_mat.solve(eig_mat @ test_v)  # cannot test equality because we will not get it

    test_V = np.ones((225, 89))
    eig_mat.solve(eig_mat @ test_V)

    # Test sampling
    samples = np.random.default_rng(2027).standard_normal(
        size=(4, 70, eig_mat._subspace_size)
    )
    np.testing.assert_allclose(
        samples, eig_mat.whiten(eig_mat.colorize(samples)), rtol=1e-2, atol=1e-2
    )


def test_negative_eigen_values() -> None:
    # we build a matrix with negative eigen values
    # matrix 4 x 4
    U = np.arange(16).reshape((4, 4)).astype(np.float64)
    V = np.diag([5.0, 4.0, -1.0, -2.0])

    # This is the dense matrix to decompose
    cov_mat = U @ V @ U.T

    cov_mat_eigen = covmats.eigen_factorize_cov_mat(cov_mat, n_pc=3, random_state=2023)
    assert cov_mat_eigen.n_pc == 2
    assert cov_mat_eigen.eig_vects.size == 8


def test_CovViaEnsemble_0() -> None:

    rng = np.random.default_rng(2026)
    n = 5
    A = rng.random(size=(n, n))
    A = A @ A.T  # make the covariance symmetric positive definite

    L = np.linalg.cholesky(A)
    cov_cho = covmats.CovViaCholesky(L)

    # Test colorize with an ensemble
    colored_samples = cov_cho.sample_mvnormal(
        shape=(5000,), random_state=np.random.default_rng(2027)
    )
    cov_ens = covmats.CovViaEnsemble(colored_samples)

    # Check to dense
    np.testing.assert_allclose(cov_ens.todense(), A, rtol=0.02)

    expected = np.array([20.926541, 33.896138, 43.155498, 25.852243, 22.575929])

    # matvec and rmatvec
    np.testing.assert_allclose(cov_ens @ v5, expected, rtol=1e-6)
    np.testing.assert_allclose(cov_ens @ v5, cov_ens.T @ v5)

    # matmat
    np.testing.assert_allclose(cov_ens @ V54, np.vstack([expected] * 4).T, rtol=1e-6)

    # rmatmat
    np.testing.assert_allclose(cov_ens @ V54, cov_ens.T @ V54)

    # solve
    np.testing.assert_allclose(cov_ens.solve(cov_ens @ v5), v5, atol=1e-10)
    np.testing.assert_allclose(cov_ens.solve(cov_ens @ V54), V54, atol=1e-10)

    # Get precision matrix (inverse)
    np.testing.assert_allclose(np.linalg.inv(A), cov_ens.precision, rtol=0.05)


def test_CovViaEnsemble_1() -> None:

    # 4225 x 4225 sparse precision matrix
    Q = load_precision_example_4225x()
    # 4225 x 4225 dense covariance matrix
    # C = np.linalg.inv(Q.toarray())

    cov = covmats.CovViaSparsePrecisionCholesky(
        covmats.SparseCholeskyFactor(*_get_L_D_P(Q))
    )

    # Sample from this covariance
    ne = 1000
    ensemble = cov.sample_mvnormal((ne,), random_state=np.random.default_rng(209))

    # Create a new Cov instance using the ensemble
    Ce = covmats.CovViaEnsemble(ensemble=ensemble)

    # Make it dense
    Ce.todense()
    # np.testing.assert_allclose(Ce.todense(), C, rtol=0.5, atol=1.0)

    # Invert it
    np.testing.assert_allclose(Ce.precision, Q.toarray(), rtol=1e-2, atol=2e-1)

    # Sample from this ensemble
    samples = np.random.default_rng(2027).standard_normal(size=(ne, Ce.subspace_size))
    np.testing.assert_allclose(
        samples, Ce.whiten(Ce.colorize(samples)), rtol=1e-2, atol=1e-2
    )

    # Test linop capability
    # TODO


def test_CovViaEnsemble_2() -> None:
    # precision matrix (sparse)
    Q = get_SPD_sparse_n11_example(seed=2026)
    # dense covariance matrix
    C = np.linalg.inv(Q.toarray())
    cvspc = covmats.CovViaSparsePrecisionCholesky(
        covmats.SparseCholeskyFactor(*_get_L_D_P(Q))
    )

    # White noise sampling
    samples = np.random.default_rng(2027).standard_normal(size=(50_000, 11))
    # Sample from the covariance matrix
    colored_samples = cvspc.colorize(samples)

    # New representation from the ensemble
    Ce = covmats.CovViaEnsemble(colored_samples)
    # Test sampling using the new representation
    colored_samples2 = Ce.colorize(samples)
    np.testing.assert_allclose(samples, Ce.whiten(colored_samples2))

    # Make a second covariance instancefrom the new ensemble
    Ce2 = covmats.CovViaEnsemble(colored_samples2)

    # Test the diagonal extraction
    np.testing.assert_allclose(np.diag(C), Ce.get_diagonal(), rtol=0.05)

    # Check consistency
    np.testing.assert_allclose(C, Ce.todense(), atol=0.05)
    np.testing.assert_allclose(C, Ce2.todense(), atol=0.1)
    np.testing.assert_allclose(Q.toarray(), Ce.precision, atol=0.05)
    np.testing.assert_allclose(Q.toarray(), Ce2.precision, atol=0.05)

    # Test linop capability
    # TODO



"""Test the priors implementation."""

import re
from contextlib import nullcontext as does_not_raise
from typing import Optional

import numpy as np
import pytest
from covmats import (
    ConstantDriftMatrix,
    ConstantPriorTerm,
    DriftMatrix,
    EnsembleMeanPriorTerm,
    LinearDriftMatrix,
    MeanPriorTerm,
    NullPriorTerm,
)


def test_null_prior() -> None:
    prior = NullPriorTerm()
    assert prior.get_values(np.ones(45)) == 0.0
    np.testing.assert_equal(prior.get_gradient_dot_product(np.ones(45)), np.zeros(45))


def test_constant_prior() -> None:
    prior_values = np.ones(22) * 5.0
    prior = ConstantPriorTerm(prior_values)

    np.testing.assert_array_equal(prior.get_values(np.ones(22) * 8.3), prior_values)

    with pytest.raises(
        ValueError,
        match=(
            r"The given values have shape \(45,\) while the "
            r"constant prior has been defined with shape \(22,\)!"
        ),
    ):
        np.testing.assert_array_equal(prior.get_values(np.ones(45) * 9.6), prior_values)

    np.testing.assert_equal(prior.get_gradient_dot_product(np.ones(45)), np.zeros(45))


def test_mean_prior() -> None:
    prior = MeanPriorTerm()

    np.testing.assert_allclose(prior.get_values(np.ones(45) * 9.6), np.ones(45) * 9.6)
    np.testing.assert_allclose(prior.get_values(np.ones(22) * 8.3), np.ones(22) * 8.3)

    np.testing.assert_allclose(prior.get_gradient_dot_product(np.ones(45)), np.ones(45))


@pytest.mark.parametrize(
    "shape,ensemble,expected_values,vector,expected_exception",
    [
        (
            (10,),
            np.array([]),
            0.0,
            np.array([]),
            pytest.raises(
                ValueError,
                match=re.escape(
                    "The shape of an EnsembleMeanPriorTerm should be (N_s, N_e)"
                    " with N_s the number of adjuted values and N_e the number of"
                    " members in the ensemble."
                ),
            ),
        ),
        (
            (10, 20),
            np.ones((2, 20)),
            0.0,
            np.array([]),
            pytest.raises(
                ValueError, match=re.escape(r"Expected shape (10, 20), got (2, 20).")
            ),
        ),
        (
            (10, 20),
            np.ones((10, 20)) * 2.0,
            np.ones((10, 1)) * 2.0,
            np.ones((10)),
            does_not_raise(),
        ),
        (
            (10, 20),
            np.ones((10, 20)) * 2.0,
            np.ones((10, 1)) * 2.0,
            np.ones(5),
            pytest.raises(
                ValueError, match=re.escape("Expected a vector of size 10, got (5,).")
            ),
        ),
    ],
)
def test_ensemble_mean_prior(
    shape, ensemble, expected_values, vector, expected_exception
) -> None:
    with expected_exception:
        prior = EnsembleMeanPriorTerm(shape)
        prior.shape == shape
        np.testing.assert_array_equal(prior.get_values(ensemble), expected_values)
        assert np.shape(prior.get_gradient_dot_product(vector)) == vector.shape


def test_drift_matrix() -> None:
    dmat = DriftMatrix(
        np.array([[2.0, 3.0, 2.0], [2.0, 3.0, 2.0]]), beta=np.array([2.0, 2.0, 2.0])
    )
    assert dmat.mat.shape == (2, 3)
    assert dmat.s_dim == 2
    assert dmat.beta_dim == 3

    np.testing.assert_allclose(dmat.get_values(np.ones(2)), np.ones(2) * 14.0)

    with pytest.raises(
        ValueError,
        match=(
            r"The given values have size 3 while the X matrix "
            r"has been defined with shape \(2, 3\)!"
        ),
    ):
        np.testing.assert_allclose(dmat.get_values(np.ones(3)), np.ones(3))

    assert dmat.get_gradient_dot_product(np.ones(45)) == 0.0


def test_drift_matrix_no_beta() -> None:
    dmat = DriftMatrix(np.array([[2.0, 3.0, 2.0], [2.0, 3.0, 2.0]]))
    with pytest.raises(ValueError, match=r"beta is None! A value must be given."):
        np.testing.assert_allclose(dmat.get_values(np.ones(2)), np.ones(2) * 14.0)


def test_drift_matrix_wrong_beta() -> Optional[DriftMatrix]:
    with pytest.raises(
        ValueError,
        match=(
            r"beta has shape \(1,\) while it should be "
            r"shape \(3,\) to match the given coefficient matrix."
        ),
    ):
        return DriftMatrix(np.array([[2.0, 3.0, 2.0], [2.0, 3.0, 2.0]]), beta=2.0)

    with pytest.raises(
        ValueError,
        match=(
            r"beta has shape \(2,\) while it should be shape \(3,\) "
            r"to match the given coefficient matrix."
        ),
    ):
        return DriftMatrix(
            np.array([[2.0, 3.0, 2.0], [2.0, 3.0, 2.0]]), beta=np.array([2.0, 2.0])
        )


def test_constant_drift_matrix() -> None:
    dmat = ConstantDriftMatrix(n_pts=5)
    assert dmat.mat.shape == (5, 1)

    dmat = ConstantDriftMatrix(n_pts=4)
    assert dmat.mat.shape == (4, 1)

    assert dmat.get_gradient_dot_product(np.ones(45)) == 0.0


def test_linear_drift_matrix() -> None:
    pts = np.array([[1.0, 1.0], [2.0, 1.0], [3.0, 1.0], [4.0, 1.0], [5.0, 1.0]])
    dmat = LinearDriftMatrix(pts)
    assert dmat.mat.shape == (5, 3)

    pts = np.array([[1.0, 1.0, 5.0], [2.0, 1.0, 5.0]])
    dmat = LinearDriftMatrix(pts)
    assert dmat.mat.shape == (2, 4)

    assert dmat.get_gradient_dot_product(np.ones(45)) == 0.0


# def test_linear_drift_matrix2() -> None:
#     nx = 100
#     ny = 100
#     grid = RectilinearGrid(
#         x0=0.0, y0=0.0, z0=0.0, dx=5.0, dy=5.0, dz=1.0, nx=nx, ny=ny, nz=1
#     )
#     # coordinates 2d
#     coords = grid.center_coords_2d.reshape(2, -1, order="F")

#     # Linear trend: b0 + b1 * x + b2 * y
#     trend = DriftMatrix(
#         np.concatenate([np.ones((1, nx * ny)), coords]).T,
#         beta=np.array([219, 0.1, -0.1]),
#     )
#     trend.get_values(np.ones(nx * ny))

#     assert trend.s_dim == nx * ny



import numpy as np
import pytest
from covmats._helpers import get_pts_coords_regular_grid
from covmats._toeplitz import (
    create_toepliz_first_row,
    toeplitz_product,
)
from covmats._types import NDArrayFloat


def exponential_kernel(r: float) -> NDArrayFloat:
    """Test covariance kernel."""
    return (150**2) * np.exp(-r)


@pytest.mark.parametrize(
    "mesh_dim, shape, x",
    [
        (
            1,
            5,
            np.ones(5) * 4.0,
        ),
        (
            (1, 7),
            (5, 5),
            np.ones(25) * 4.0,
        ),
        (
            (1, 7),
            (5, 5),
            np.ones(25) * 4.0,
        ),
        (
            (2, 7, 8),
            np.array((5, 5, 8)),
            np.ones(200) * 4.0,
        ),
    ],
)
def test_toeplitz(mesh_dim, shape, x) -> None:
    pts = get_pts_coords_regular_grid(mesh_dim, shape)
    first_row = create_toepliz_first_row(pts, exponential_kernel, shape)
    toeplitz_product(x, first_row, shape)


def test_toeplitz_product_exception() -> None:
    with pytest.raises(ValueError, match="Support 1,2 and 3 dimensions"):
        toeplitz_product(np.ones(16 * 3), np.ones(16 * 3), (1, 3, 4, 4))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Antoine COLLET

from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

import numpy as np

from covmats._types import NDArrayFloat


class PriorTerm(ABC):
    """Represent a prior term for the geostatistical regularization."""

    @abstractmethod
    def get_values(self, params: NDArrayFloat) -> Union[float, NDArrayFloat]:
        """
        Return the values of the prior term.

        Parameters
        ----------
        params : NDArrayFloat
            Values of the parameters for which to compute the prior.

        Returns
        -------
        NDArrayFloat
            The prior term values.
        """

    @abstractmethod
    def get_gradient_dot_product(
        self, input: NDArrayFloat
    ) -> Union[float, NDArrayFloat]:
        """
        Return the dot product of the gradient of the prior and the given input vector.

        Parameters
        ----------
        params : NDArrayFloat
            Values with which to compute the prior gradient dot product.

        Returns
        -------
        NDArrayFloat
            Prior gradient-input vector dot product.
        """


class NullPriorTerm(PriorTerm):
    """Represent a null prior term."""

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def get_values(self, params: NDArrayFloat) -> float:
        """
        Return the values of the prior term.

        Parameters
        ----------
        params : NDArrayFloat
            Values of the parameters for which to compute the prior. It has no effect
            with `NullPriorTerm`.

        Returns
        -------
        NDArrayFloat
            The prior term values.
        """
        return 0.0

    def get_gradient_dot_product(
        self, input: NDArrayFloat
    ) -> Union[float, NDArrayFloat]:
        """
        Return the dot product of the gradient of the prior and the given input vector.

        Parameters
        ----------
        params : NDArrayFloat
            Values with which to compute the prior gradient dot product.

        Returns
        -------
        NDArrayFloat
            Prior gradient-input vector dot product.
        """
        return np.zeros(input.shape)


class ConstantPriorTerm(PriorTerm):
    """Represent a prior (no influence of beta)."""

    def __init__(self, prior_values: NDArrayFloat) -> None:
        """
        Initialize the instance.

        Parameters
        ----------
        prior_values : NDArrayFloat
            Values given to the prior term.
        """
        super().__init__()
        self.prior_values: NDArrayFloat = prior_values.ravel("F")

    def get_values(self, params: NDArrayFloat) -> NDArrayFloat:
        """
        Return the values of the prior term.

        Parameters
        ----------
        params : NDArrayFloat
            Values of the parameters for which to compute the prior. It has no effect
            with `ConstantPriorTerm`.

        Returns
        -------
        NDArrayFloat
            The prior term values.
        """
        if params.shape != self.prior_values.shape:
            raise ValueError(
                f"The given values have shape {params.shape} while the constant prior "
                f"has been defined with shape {self.prior_values.shape}!"
            )
        return self.prior_values

    def get_gradient_dot_product(
        self, input: NDArrayFloat
    ) -> Union[float, NDArrayFloat]:
        """
        Return the dot product of the gradient of the prior and the given input vector.

        Parameters
        ----------
        params : NDArrayFloat
            Values with which to compute the prior gradient dot product.

        Returns
        -------
        NDArrayFloat
            Prior gradient-input vector dot product.
        """
        return np.zeros(input.shape)


class MeanPriorTerm(PriorTerm):
    """Represent a mean prior."""

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def get_values(self, params: NDArrayFloat) -> NDArrayFloat:
        """
        Return the values of the prior term.

        Parameters
        ----------
        params : NDArrayFloat
            Values of the parameters for which to compute the prior mean.

        Returns
        -------
        NDArrayFloat
            The prior term values.
        """
        return np.full(params.size, fill_value=np.sum(params)) / params.size

    def get_gradient_dot_product(
        self, input: NDArrayFloat
    ) -> Union[float, NDArrayFloat]:
        """
        Return the dot product of the gradient of the prior and the given input vector.

        Parameters
        ----------
        params : NDArrayFloat
            Values with which to compute the prior gradient dot product.

        Returns
        -------
        NDArrayFloat
            Prior gradient-input vector dot product.
        """
        return np.full(input.size, fill_value=np.sum(input)) / input.size


class EnsembleMeanPriorTerm(PriorTerm):
    """Represent a mean prior."""

    def __init__(self, shape: Tuple[int, ...]) -> None:
        """Initialize the instance."""
        super().__init__()
        if len(shape) != 2:
            raise ValueError(
                "The shape of an EnsembleMeanPriorTerm should be (N_s, N_e)"
                " with N_s the number of adjuted values and N_e the number of"
                " members in the ensemble."
            )
        self.shape: Tuple[int, ...] = shape

    def get_values(self, params: NDArrayFloat) -> NDArrayFloat:
        """
        Return the values of the prior term.

        Parameters
        ----------
        params : NDArrayFloat
            Values of the parameters for which to compute the prior mean.

        Returns
        -------
        NDArrayFloat
            The prior term values.
        """
        if params.shape != self.shape:
            raise ValueError(f"Expected shape {self.shape}, got {params.shape}.")
        return np.mean(params, axis=1, keepdims=True)

    def get_gradient_dot_product(
        self, input: NDArrayFloat
    ) -> Union[float, NDArrayFloat]:
        """
        Return the dot product of the gradient of the prior and the given input vector.

        Parameters
        ----------
        params : NDArrayFloat
            Values with which to compute the prior gradient dot product.

        Returns
        -------
        NDArrayFloat
            Prior gradient-input vector dot product.
        """
        if input.shape[0] != self.shape[0]:
            raise ValueError(
                f"Expected a vector of size {self.shape[0]}, got {input.shape}."
            )
        return input / self.shape[1]


class DriftMatrix(PriorTerm):
    """Represent a drift matrix prior term."""

    __slots__: List[str] = ["mat"]

    def __init__(
        self, mat: NDArrayFloat, beta: Optional[Union[NDArrayFloat, float]] = None
    ) -> None:
        """_summary_

        Parameters
        ----------
        mat : NDArrayFloat
            Matrix of coefficients: X. with shape (Ns, Nbeta)
        beta : Optional[Union[NDArrayFloat, float]], optional
            P Coefficients, by default None. # TODO: add references and comment better.
        """
        self.mat: NDArrayFloat = mat
        self.beta: Optional[Union[NDArrayFloat, float]] = beta

        if beta is not None:
            if isinstance(beta, float):
                shape = (1,)
            else:
                shape = np.shape(beta)
            if shape[0] != mat.shape[1]:
                raise ValueError(
                    f"beta has shape {shape} while it should be shape "
                    f"({mat.shape[1]},) to match the given coefficient matrix."
                )

    @property
    def s_dim(self) -> int:
        return self.mat.shape[0]

    @property
    def beta_dim(self) -> int:
        return self.mat.shape[1]

    def dot(self, beta: Union[float, NDArrayFloat]) -> NDArrayFloat:
        """Return the dot product."""
        return np.dot(self.mat, beta)

    def get_values(self, params: NDArrayFloat) -> NDArrayFloat:
        """
        Return the values of the prior term.

        Parameters
        ----------
        params : NDArrayFloat
            Values of the parameters for which to compute the prior. It has no effect
            with `DriftMatrix`.

        Returns
        -------
        NDArrayFloat
            The prior term values.
        """
        if params.size != self.mat.shape[0]:
            raise ValueError(
                f"The given values have size {params.size} while the X matrix "
                f"has been defined with shape {self.mat.shape}!"
            )
        if self.beta is None:
            raise ValueError("beta is None! A value must be given.")
        return self.dot(self.beta)

    def get_gradient_dot_product(
        self, input: NDArrayFloat
    ) -> Union[float, NDArrayFloat]:
        """
        Return the dot product of the gradient of the prior and the given input vector.

        Parameters
        ----------
        params : NDArrayFloat
            Values with which to compute the prior gradient dot product.

        Returns
        -------
        NDArrayFloat
            Prior gradient-input vector dot product.
        """
        return 0.0


class ConstantDriftMatrix(DriftMatrix):
    """Represent a constant drift matrix (trend)."""

    # TODO: complete this one and complexify a bit

    def __init__(self, n_pts: int) -> None:
        """_summary_

        Parameters
        ----------
        pts : NDArrayFloat
            _description_
        """
        mat: NDArrayFloat = np.ones((n_pts, 1), dtype="d") / np.sqrt(n_pts)
        super().__init__(mat)


class LinearDriftMatrix(DriftMatrix):
    """Represent a linear drift matrix (trend)."""

    # TODO: complete this one and complexify a bit

    def __init__(self, pts: NDArrayFloat) -> None:
        """_summary_

        Parameters
        ----------
        pts : NDArrayFloat
            _description_
        """
        mat: NDArrayFloat = np.ones((pts.shape[0], 1 + pts.shape[1]), dtype=np.float64)
        mat[:, 1 : mat.shape[1]] = np.copy(pts)
        super().__init__(mat)



import re

import covmats
import numpy as np
import pytest
import scipy as sp
from covmats import (
    SparseCholeskyFactor,
    load_precision_example_4225x,
    load_precision_example_4225x_SCF,
)
from covmats._sparse_helpers import (
    assert_allclose_sparse,
    get_SPD_sparse_example,
    get_SPD_sparse_n11_example,
)

from .sparse_helpers import _get_L_D_P  # ty:ignore[unresolved-import]


def test_SparseCholeskyFactor_construtor() -> None:

    A = get_SPD_sparse_n11_example(seed=2026)
    L, D, P = _get_L_D_P(A)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "`L` has shape (11, 11), "
            "`D` has shape (11, 11), and `L` has shape (10,)!\n"
            "To be consistent, `D` and "
            "`P` are expected with shape (11, 11) and (11,) respectively."
        ),
    ):
        SparseCholeskyFactor(L, D, P[:-1])

    with pytest.raises(
        ValueError,
        match=("The diagonal elements of `L` are assumed to be 1!"),
    ):
        L[0, 0] = 3.0  # no unit diagonal
        SparseCholeskyFactor(L, D, P)

    with pytest.raises(
        ValueError,
        match=("The input `L` must be a lower-triangular sparse array."),
    ):
        L = L @ L.T  # not a lower triangle
        SparseCholeskyFactor(L, D, P)


def test_SparseCholeskyFactor() -> None:

    A = get_SPD_sparse_n11_example(seed=2026)
    Q = np.linalg.inv(A.toarray())
    scf = SparseCholeskyFactor(*_get_L_D_P(A))

    # Test to dense
    np.testing.assert_allclose(scf.todense(), A.toarray())
    # Test shape
    assert scf.shape == (11, 11)
    assert scf.n_pts == 11

    # Diagonal
    np.testing.assert_allclose(scf.get_diagonal(), A.diagonal())
    np.testing.assert_allclose(scf.get_invdiagonal(), Q.diagonal())

    # Test solve
    expected_x = np.arange(scf.n_pts, dtype=np.float64)
    b = A @ expected_x
    np.allclose(scf.solve(b), expected_x)

    # Test colorize and whiten
    np.testing.assert_allclose(
        scf.whiten(scf.colorize(np.eye(11))), np.eye(11), atol=1e-15
    )

    # Test colorize with an ensemble
    colored_samples = scf.colorize(
        np.random.default_rng(2027).standard_normal(size=(500_000, 11))
    )
    np.testing.assert_allclose(
        covmats.CovViaEnsemble(colored_samples).todense(), A.toarray(), atol=2e-2
    )

    # slogdet
    np.testing.assert_allclose(np.linalg.slogdet(A.toarray())[1], scf.log_pdet)


def test_assert_allclose_sparse() -> None:
    L = sp.sparse.csc_array([[1.0, 0.0, 0.1], [0.1, 1.2, 0.0], [0.0, 0.0, 1.5]])
    A = L.T @ L

    # works
    assert_allclose_sparse(A, A)
    assert_allclose_sparse(A.T, A)
    assert_allclose_sparse(L, L)

    # shapes are not the same
    with pytest.raises(AssertionError):
        assert_allclose_sparse(L, L.T)

    # Values are not the same
    with pytest.raises(AssertionError):
        assert_allclose_sparse(A, L)

    # shape is not the same
    with pytest.raises(AssertionError):
        assert_allclose_sparse(
            A, sp.sparse.csc_array([[1.0, 0.0, 0.1], [0.0, 0.0, 1.5]])
        )

    # empty
    assert_allclose_sparse(sp.sparse.csc_array([[]]), sp.sparse.csc_array([[]]))


def test_get_SPD_sparse_example() -> None:

    A = get_SPD_sparse_example(100, 2025, diag_mean=4.0)
    SparseCholeskyFactor(*_get_L_D_P(A))


def test_load_precision_example_4225x() -> None:

    A = load_precision_example_4225x()
    SparseCholeskyFactor(*_get_L_D_P(A))

    # Load the cholesky factor
    load_precision_example_4225x_SCF()


def test_indefinite():

    A = get_SPD_sparse_example(100, 2025, diag_mean=5.0)
    L, D, P = _get_L_D_P(A)
    D = sp.sparse.diags(np.hstack([np.ones(100), 0.0]))

    # And sparse cholesky should fail as well
    with pytest.raises(
        ValueError,
        match="1 null values have been detected in `D` which means the factorized"
        " matrix is singular and non invertible.",
    ):
        SparseCholeskyFactor(L, D, P)


def test_singular():

    # Produce a singular matrix
    A = get_SPD_sparse_example(100, 2025, diag_mean=0.0)
    # Compute the eigenvalues of a complex Hermitian or real symmetric matrix
    sigma = np.linalg.eigvalsh(A.toarray())
    # There are 53 negative eigen values
    assert np.count_nonzero(sigma < 0) == 53

    # classic cholesky fails
    with pytest.raises(np.linalg.LinAlgError):
        sp.linalg.cholesky(A.toarray())

    # And sparse cholesky should fail as well
    with pytest.raises(
        ValueError,
        match="53 negative values have been detected in `D` which means the factorized"
        " matrix is indefinite and non invertible.",
    ):
        SparseCholeskyFactor(*_get_L_D_P(A))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Antoine COLLET

"""Provide some helpers."""

from typing import Sequence, Union

import numpy as np
from scipy._lib._util import check_random_state as check_random_state

from covmats._types import NDArrayFloat, NDArrayInt


def get_pts_coords_regular_grid(
    mesh_dim: Union[float, Sequence[float], NDArrayFloat],
    shape: Union[int, Sequence[int], NDArrayInt],
) -> NDArrayFloat:
    """
    Create an array of points coordinates for regular grids.

    It supports from 1 to n dimensions.

    Parameters
    ----------
    mesh_dim : NDArrayInt
        Dimensions of one mesh of the grid.
    shape : NDArrayInt
        Shape of the grid (number of grid cells along each axis). The number of elements
        in `shape` much match the number
        of elements in `mesh_dim`.

    Returns
    -------
    NDArrayFloat
        Array of coordinates with shape (Npts, Ndims).
    """
    # convert to numpy array
    _mesh_dim = np.array([mesh_dim]).ravel()
    _shape = np.array(shape, dtype=np.int64).ravel()
    # xmin = center of the first mesh
    xmin: NDArrayFloat = np.array(_mesh_dim) / 2.0
    # xmax  = center of the last mesh
    xmax = (_shape - 0.5) * _mesh_dim
    return (
        np.array(
            np.meshgrid(
                *[np.linspace(xmin[i], xmax[i], _shape[i]) for i in range(_shape.size)],
                indexing="ij",
            )
        )
        .reshape(_shape.size, -1, order="F")
        .T
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Antoine COLLET

"""Covariance matrix representations and priors for large scale inverse problems.

Abstract Covariance class
^^^^^^^^^^^^^^^^^^^^^^^^^

The common interface `CovarianceMatrix
<https://covmats.readthedocs.io/en/latest/_autosummary/covmats.CovarianceMatrix.html#covmats.CovarianceMatrix>`_
can be seen as a extension of the class `scipy.stats.Covariance
<https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.Covariance.html>`_
as it inherits from it (thus making it compatible with all
`scipy.stats <https://docs.scipy.org/doc/scipy/reference/stats.html>`_
functions and classes)
and dope it with `LinearOperator
<https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.LinearOperator.html>`_
capabilities.

.. autosummary::
   :toctree: _autosummary

    CovarianceMatrix


Full-rank covariance matrix decomposition
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Several full-rank representation of covariance matrices.

.. autosummary::
   :toctree: _autosummary

    CovViaDiagonal
    CovViaCholesky
    CovViaSparseCholesky
    CovViaPrecisionCholesky
    CovViaSparsePrecisionCholesky

Low-rank approximations
^^^^^^^^^^^^^^^^^^^^^^^

Several low-rank (approximate) representation of covariance matrices.

.. autosummary::
   :toctree: _autosummary

    CovViaEigenFactorization
    CovViaEnsemble

Kernel based approximations
^^^^^^^^^^^^^^^^^^^^^^^^^^^

Low-rank approximations from a given Kernel. Only provides the linear operations
capabilities (no sampling not statistical calculations).

.. autosummary::
   :toctree: _autosummary

    CovKernelAsLinop
    CovKernelAsLinopViaFFT

Matrix compression
^^^^^^^^^^^^^^^^^^^

Allow to factorize :py:class:`CovarianceMatrix` and :py:class:`CovKernelAsLinop` using
Eigen low-rank approximation.

.. autosummary::
   :toctree: _autosummary

    get_linop_eigen_factorization
    eigen_factorize_cov_mat

Sparse Helpers
^^^^^^^^^^^^^^

Helper to work with sparse matrices and covariances.

.. autosummary::
   :toctree: _autosummary

    SparseCholeskyFactor

Working with priors and trends
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

To represent trend through drift matrix. To use along with geostatistical regularizator.

.. autosummary::
   :toctree: _autosummary

    PriorTerm
    NullPriorTerm
    ConstantPriorTerm
    MeanPriorTerm
    EnsembleMeanPriorTerm
    DriftMatrix
    ConstantDriftMatrix
    LinearDriftMatrix

Other utility functions
^^^^^^^^^^^^^^^^^^^^^^^

To work with covariance matrices and low rank approximations.

.. autosummary::
   :toctree: _autosummary

    get_explained_var

Test data
^^^^^^^^^
Functions providing test data.

.. autosummary::
   :toctree: _autosummary

    load_precision_example_4225x
    load_precision_example_4225x_SCF
    get_SPD_sparse_n11_example
    get_SPD_sparse_example


"""

from covmats.__about__ import __author__, __email__, __version__
from covmats._covariances import (
    CovarianceMatrix,
    CovKernelAsLinop,
    CovKernelAsLinopViaFFT,
    CovViaCholesky,
    CovViaDiagonal,
    CovViaEigenFactorization,
    CovViaEnsemble,
    CovViaPrecisionCholesky,
    CovViaSparseCholesky,
    CovViaSparsePrecisionCholesky,
    eigen_factorize_cov_mat,
    get_explained_var,
    get_linop_eigen_factorization,
)
from covmats._priors import (
    ConstantDriftMatrix,
    ConstantPriorTerm,
    DriftMatrix,
    EnsembleMeanPriorTerm,
    LinearDriftMatrix,
    MeanPriorTerm,
    NullPriorTerm,
    PriorTerm,
)
from covmats._sparse_helpers import (
    SparseCholeskyFactor,
    get_SPD_sparse_example,
    get_SPD_sparse_n11_example,
)
from covmats.data import (
    load_precision_example_4225x,
    load_precision_example_4225x_SCF,
)

__all__ = [
    "CovarianceMatrix",
    "CovViaDiagonal",
    "CovViaCholesky",
    "CovViaSparseCholesky",
    "CovViaPrecisionCholesky",
    "CovViaSparsePrecisionCholesky",
    "CovViaEigenFactorization",
    "CovViaEnsemble",
    "CovKernelAsLinopViaFFT",
    "CovKernelAsLinop",
    "PriorTerm",
    "NullPriorTerm",
    "ConstantPriorTerm",
    "MeanPriorTerm",
    "EnsembleMeanPriorTerm",
    "DriftMatrix",
    "ConstantDriftMatrix",
    "LinearDriftMatrix",
    "get_explained_var",
    "get_linop_eigen_factorization",
    "eigen_factorize_cov_mat",
    "__version__",
    "__email__",
    "__author__",
    "load_precision_example_4225x",
    "load_precision_example_4225x_SCF",
    "SparseCholeskyFactor",
    "get_SPD_sparse_n11_example",
    "get_SPD_sparse_example",
]



from importlib import resources

import numpy as np
import scipy as sp

from covmats._sparse_helpers import SparseCholeskyFactor


def load_precision_example_4225x() -> sp.sparse.csc_array:
    with (
        resources.files("covmats.data")
        .joinpath("precision_example_4225x.mtx")
        .open("rb")
    ) as f:
        return sp.io.mmread(f).tocsc()


def load_precision_example_4225x_SCF() -> SparseCholeskyFactor:
    L = (
        resources.files("covmats.data")
        .joinpath("precision_example_4225x_L.mtx")
        .open("rb")
    )
    D = (
        resources.files("covmats.data")
        .joinpath("precision_example_4225x_D.mtx")
        .open("rb")
    )
    P = np.loadtxt(
        resources.files("covmats.data").joinpath("precision_example_4225x_P.txt")  # ty:ignore[invalid-argument-type]
    )
    return SparseCholeskyFactor(sp.io.mmread(L).tocsc(), sp.io.mmread(D).tocsc(), P)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Antoine COLLET

"""Wrap the sparse cholesky factorization from"""

import warnings
from typing import Tuple, TypeVar, overload

import numpy as np
import scipy as sp

from covmats._types import ArrayLike, NDArrayFloat, NDArrayInt

# Define a type variable for the input/output type
T = TypeVar("T", NDArrayFloat, sp.sparse.csc_array)

_fact_ds = (
    r":math:`\mathbf{LDL}^{\mathrm{T}} = \mathbf{PAP}^{\mathrm{T}}` factorization"
)


class SparseCholeskyFactor(sp.sparse.linalg.LinearOperator):
    r"""
    Sparse Cholesky factor of a matrix :math:`\mathbf{A}`.

    .. _factref:

    .. admonition:: Factorization

        It relies on the :math:`\mathbf{LDL}^{\mathrm{T}} =`
        :math:`\mathbf{PAP}^{\mathrm{T}}` factorization where

        - :math:`\mathbf{L}` is the sparse lower triangle with unit diagonal.
        - :math:`\mathbf{D}` is the sparse diagonal matrix with diagonal entries.
        - :math:`\mathbf{P}` is the rows permutation vector.

    Note
    ----
    Since this code is released under BSD 3-Clause License, it is not possible to
    include sparse factors provided by
    `scikit-sparse <https://github.com/scikit-sparse/scikit-sparse>`_,
    as it depends on external libraries with GPL licenses, such as
    `SuiteSparse <https://github.com/DrTimothyAldenDavis/SuiteSparse?tab=License-1-ov-file>`_.
    The main goal of this implementation is therefore to mimic useful features while
    preserving the BSD 3-Clause License compatibility.

    Examples
    --------
    >>> from your_module import SparseCholeskyFactor
    >>> L, D, P = ...  # sparse matrices
    >>> factor = SparseCholeskyFactor(L, D, P)
    >>> factor.matvec(x)
    array([...])
    """

    __slots__ = ["_P", "_Pt", "_D", "_invD", "_sqrtD", "_sqrtinvD", "_L", "_n_pts"]

    def __init__(
        self, L: sp.sparse.sparray, D: sp.sparse.sparray, P: ArrayLike
    ) -> None:
        r"""
        Initialize the instance.

        Parameters
        ----------
        L : sp.sparse.sparray
            Sparse lower triangle :math:`\mathbf{L}` of
            the :ref:`factorization <factref>`.
        D : sp.sparse.sparray
            Sparse diagonal matrix $\mathbf{{D}}$ of the :ref:`factorization <factref>`.
        P : ArrayLike
            1D vector $\mathbf{{P}}$ storing rows permutations of the
            :ref:`factorization <factref>`.

        Raises
        ------
        ValueError
            If the dimensions of `L`, `D` and `P` do not match.
        """
        # Make sure that is it 1D and integer
        self.P = np.asarray(P, dtype=np.int64).ravel()
        self.D = D
        self.L = L

        # Assert shapes are ok.
        if self.D.shape != self.shape or self.P.size != self.n_pts:
            raise ValueError(
                f"`L` has shape {self.shape},"
                f" `D` has shape {D.shape}, and `L` has shape {self.P.shape}!\n"
                f"To be consistent, `D` and `P` are expected with shape {self.shape}"
                f" and ({self.n_pts},) respectively."
            )

    @staticmethod
    def _validate_lower_triangle(A: sp.sparse.sparray, name: str) -> sp.sparse.sparray:
        """Assert that the input matrix A is lower triangular."""
        try:
            assert_allclose_sparse(A, sp.sparse.tril(A))
        except AssertionError as e:
            message = f"The input `{name}` must be a lower-triangular sparse array."
            raise ValueError(message) from e
        return A

    @property
    def L(self) -> sp.sparse.csc_array:
        r"""
        Sparse lower triangle :math:`\mathbf{L}` with unit diagonal of
        the :ref:`factorization <factref>`.
        """
        return self._L

    @L.setter
    def L(self, L: sp.sparse.csc_array) -> None:
        r"""
        Set the sparse lower triangle of
        the :ref:`factorization <factref>`
        """
        self._L = self._validate_lower_triangle(L, "L")
        # Extract the diagonal
        try:
            np.testing.assert_allclose(L.diagonal(), np.ones(L.shape[0]))
        except AssertionError as e:
            raise ValueError("The diagonal elements of `L` are assumed to be 1!") from e

    @property
    def D(self) -> sp.sparse.csc_array:
        r"""
        Sparse diagonal matrix :math:`\mathbf{D}` of
        the :ref:`factorization <factref>`.
        """
        return self._D

    @D.setter
    def D(self, D: sp.sparse.csc_array) -> None:
        r"""
        Set the sparse diagonal matrix :math:`\mathbf{D}` of
        the :ref:`factorization <factref>`.
        """
        nnv = np.count_nonzero(D.diagonal() < 0.0)
        if nnv != 0:
            raise ValueError(
                f"{nnv} negative values have been detected in `D` which "
                "means the factorized matrix is indefinite and non invertible."
            )

        nnv = np.count_nonzero(D.diagonal() == 0.0)
        if nnv != 0:
            raise ValueError(
                f"{nnv} null values have been detected in `D` which "
                "means the factorized matrix is singular and non invertible."
            )

        self._D = D
        # Update the pseudo inverse as well
        self._invD = sp.sparse.diags(1.0 / self.D.diagonal())
        # Store the square roots for whitening and colorizing transformations
        self._sqrtD = np.sqrt(self.D)
        self._sqrtinvD = np.sqrt(self.invD)

    @property
    def invD(self) -> sp.sparse.csc_array:
        r"""
        Inverse of :math:`\mathbf{D}` in
        the :ref:`factorization <factref>`.

        Notes
        -----
        The property is read-only and updated when setting `D`.
        """
        return self._invD

    @property
    def sqrtinvD(self) -> sp.sparse.csc_array:
        r"""
        Squared root inverse of :math:`\mathbf{D}` in
        the :ref:`factorization <factref>`.

        Notes
        -----
        The property is read-only and updated when setting `D`.
        """
        return self._sqrtinvD

    @property
    def sqrtD(self) -> sp.sparse.csc_array:
        r"""
        Squared root of :math:`\mathbf{D}` in
        the :ref:`factorization <factref>`.

        Notes
        -----
        The property is read-only and updated when setting `D`.
        """
        return self._sqrtD

    @property
    def P(self) -> NDArrayInt:
        r"""
        1D vector :math:`\mathbf{P}` storing rows permutations of the
        :ref:`factorization <factref>`.
        """
        return self._P

    @P.setter
    def P(self, P: ArrayLike) -> None:
        r"""
        Set the 1D vector :math:`\mathbf{P}` storing rows permutations of the
        :ref:`factorization <factref>`.
        """
        self._P = np.asarray(P, dtype=np.int64).ravel()
        # Update the back pivot
        self._Pt: NDArrayInt = np.argsort(self._P)

    @property
    def Pt(self) -> NDArrayInt:
        r"""
        1D vector :math:`\mathbf{P}` storing rows back-permutations of the
        :ref:`factorization <factref>`.
        """
        return self._Pt

    @overload
    def apply_P(self, x: NDArrayFloat) -> NDArrayFloat: ...

    @overload
    def apply_P(self, x: sp.sparse.csc_array) -> sp.sparse.csc_array: ...

    def apply_P(self, x: T) -> T:
        """Apply the permutation of rows."""
        return x[self.P]

    @overload
    def apply_Pt(self, x: NDArrayFloat) -> NDArrayFloat: ...

    @overload
    def apply_Pt(self, x: sp.sparse.csc_array) -> sp.sparse.csc_array: ...

    def apply_Pt(self, x: T) -> T:
        """Apply the reverse permutation of rows."""
        return x[self.Pt]

    def solve(self, b: NDArrayFloat) -> NDArrayFloat:
        r"""
        Return :math:`\mathbf{x} = \mathbf{A}^{-1} \mathbf{b}.`

        # L @ D @ L' = P @ A @ P'
        # A x = b
        # P' L @ D @ L' P x = b
        # x = P' L^-T @ D^{-1} @ L^{-1} P b

        Parameters
        ----------
        b: NDArrayFloat
            Column vector with shape (n,) or ensemble matrix with shape (n, ne).

        Returns
        -------
        NDArrayFloat
            _description_
        """
        return self.apply_Pt(
            sp.sparse.linalg.spsolve_triangular(
                self.L.T,
                self.invD
                @ sp.sparse.linalg.spsolve_triangular(
                    self.L, self.apply_P(b), lower=True, unit_diagonal=True
                ),
                lower=False,
                unit_diagonal=True,
            )
        )

    def _matvec(self, x: NDArrayFloat) -> NDArrayFloat:
        """Return A @ x."""
        L = self.apply_Pt(self.L)
        return L @ (self.D @ (L.T @ x))

    @property
    def log_pdet(self) -> float:
        """
        Log of the pseudo-determinant of the covariance matrix.

        Note
        ----
        In linear algebra and statistics, the pseudo-determinant [1] is the product of
        all non-zero eigenvalues of a square matrix. It coincides with the regular
        determinant when the matrix is non-singular.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return np.log(np.prod(self.D.diagonal()))

    @property
    def n_pts(self) -> int:
        """Number of points in the domain (n). This is read-only."""
        return self.L.shape[0]

    @property
    def shape(self) -> Tuple[int, int]:
        """
        Shape of the square matrix (n, n). Read-only.
        """
        return self.L.shape[:2]

    @property
    def mat(self) -> sp.sparse.csc_array:
        r"""
        Return :math:`\mathbf{A}` as a dense array.

        Note
        ----
        Alias for :py:meth:`SparseCholeskyFactor.todense()`. This is read-only.
        """
        # Permute and rebuild
        L = self.apply_Pt(self.L)
        return L @ self.D @ L.T

    def todense(self) -> NDArrayFloat:
        r"""Return :math:`\mathbf{A}` as a dense array."""
        # Permute and rebuild
        return self.mat.toarray()

    def inv(self) -> NDArrayFloat:
        r"""Return the inverse :math:`\mathbf{A}^{-1}` as a dense array."""
        return self.solve(np.eye(self.n_pts))

    def get_diagonal(self) -> NDArrayFloat:
        r"""
        Return the diagonal entries of the matrix :math:`\mathbf{A}`.

        It as shape (n,).
        """
        return (self.L @ self.D @ self.L.T).diagonal()

    def get_invdiagonal(self) -> NDArrayFloat:
        r"""
        Return the diagonal entries of the inverse matrix :math:`\mathbf{A}^{-1}`.

        It as shape (n,).
        """
        cov_mat_diag = np.zeros(self.n_pts)
        v = np.zeros(self.n_pts)
        for i in range(self.n_pts):
            v[i - 1] = 0.0
            v[i] = 1.0
            cov_mat_diag[i] = self.solve(v)[i]
        return cov_mat_diag

    def colorize(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a colorizing transformation on data.

        "Colorizing" ("color" as in "colored noise", in which different
        frequencies may have different magnitudes) transforms a set of
        uncorrelated random variables into a new set of random variables with
        the desired covariance. When a coloring transform is applied to a
        sample of points distributed according to a multivariate normal
        distribution with identity covariance and zero mean, the covariance of
        the transformed sample is approximately the covariance matrix used
        in the coloring transform
        :cite:p:`WhiteningTransformation2025,novakGeneralizationColoringLinear2019`.

        Parameters
        ----------
        x : array_like
            An array of points. The last dimension must correspond with the
            dimensionality of the space, i.e., the number of columns in the
            covariance matrix.

        Returns
        -------
        x_ : array_like
            The transformed array of points.

        Note
        ----
        We want to solve z.T = x @ K.T, where A = K @ K^{T}
        We use the cholesky factorization LDL' = PA'AP'
        with P' = P^{-1} the permutation that makes the decomposition unique.
        So LD^{1/2} = PA' and A = D^{1/2}L'P
        Finally z = (P' L D^{1/2} x')'

        References
        ----------
        .. bibliography::
            :list: enumerated
            :filter: False

            WhiteningTransformation2025
            novakGeneralizationColoringLinear2019

        Examples
        --------
        >>> import numpy as np
        >>> import covmats
        >>> rng = np.random.default_rng(1638083107694713882823079058616272161)
        >>> n = 3
        >>> A = rng.random(size=(n, n))
        >>> cov_array = A @ A.T  # make matrix symmetric positive definite
        >>> cholesky = np.linalg.cholesky(cov_array)
        >>> cov_object = covmats.CovViaCholesky(cholesky)
        >>> x = rng.multivariate_normal(np.zeros(n), np.eye(n), size=(10000))
        >>> x_ = cov_object.colorize(x)
        >>> cov_data = np.cov(x_, rowvar=False)
        >>> np.allclose(cov_data, cov_array, rtol=3e-2)
        True
        """
        return self.apply_Pt(self.L @ self.sqrtD @ x.T).T

    def colorize_inv(self, x: NDArrayFloat) -> NDArrayFloat:
        r"""
        Perform a colorizing transformation on data as in
        :py:meth:`SparseCholeskyFactor:colorize` but for the inverse
        :math:`\mathbf{A}^{-1}`.
        """
        return self.apply_Pt(
            sp.sparse.linalg.spsolve_triangular(
                self.L.T, self.sqrtinvD @ x.T, unit_diagonal=True, lower=False
            )
        ).T

    def whiten(self, x: NDArrayFloat) -> NDArrayFloat:
        """
        Perform a whitening transformation on data.

        "Whitening" ("white" as in "white noise", in which each frequency has
        equal magnitude) transforms a set of random variables into a new set of
        random variables with unit-diagonal covariance. When a whitening
        transform is applied to a sample of points distributed according to
        a multivariate normal distribution with zero mean, the covariance of
        the transformed sample is approximately the identity matrix
        :cite:p:`WhiteningTransformation2025, novakGeneralizationColoringLinear2019`.

        Parameters
        ----------
        x : array_like
            An array of points. The last dimension must correspond with the
            dimensionality of the space, i.e., the number of columns in the
            covariance matrix.

        Returns
        -------
        x_ : array_like
            The transformed array of points.

        Note
        ----
        TODO.
        We use the cholesky factorization LDL' = PA'AP'
        We want to solve x = z @ A.T =>  z = x A^{-T}
        with P' = P^{-1} the permutation that makes the decomposition unique.
        So LD^{1/2} = PA' and A = D^{1/2}L'P
        Finally z = (D^{-1/2} L^{-1} P x')'

        References
        ----------
        .. bibliography::
            :list: enumerated
            :filter: False

            novakGeneralizationColoringLinear2019
            WhiteningTransformation2025

        Examples
        --------
        >>> import numpy as np
        >>> import covmats
        >>> rng = np.random.default_rng()
        >>> n = 3
        >>> A = rng.random(size=(n, n))
        >>> cov_array = A @ A.T  # make matrix symmetric positive definite
        >>> precision = np.linalg.inv(cov_array)
        >>> cov_object = covmats.CovViaPrecisionCholesky(
        .. sp.linalg.cholesky(precision, lower=True))
        >>> x = rng.multivariate_normal(np.zeros(n), cov_array, size=(10000))
        >>> x_ = cov_object.whiten(x)
        >>> np.cov(x_, rowvar=False)  # near-identity covariance
        array([[0.97862122, 0.00893147, 0.02430451],
               [0.00893147, 0.96719062, 0.02201312],
               [0.02430451, 0.02201312, 0.99206881]])

        """
        return (
            self.sqrtinvD
            @ sp.sparse.linalg.spsolve_triangular(
                self.L, self.apply_P(x.T), unit_diagonal=True, lower=True
            )
        ).T

    def whiten_inv(self, x: NDArrayFloat) -> NDArrayFloat:
        r"""
        Perform a whitening transformation on data as in
        :py:meth:`SparseCholeskyFactor:whiten` but for the inverse
        :math:`\mathbf{A}^{-1}`.
        """
        return self.sqrtD @ (self.L.T @ self.apply_P(x.T)).T


def assert_allclose_sparse(
    A: sp.sparse.sparray, B: sp.sparse.sparray, atol: float = 1e-8, rtol: float = 1e-8
) -> None:
    """
    Assert that two sparse matrices or arrays are almost equal.

    Parameters
    ----------
    A : sp.sparse.sparray
        First sparse matrix.
    B : sp.sparse.sparray
        Second sparse matrix.
    atol : float, optional
        Absolute tolerance for the equality test, by default 1e-8.
    rtol : float, optional
        Relative tolerance for the equality test, by default 1e-8.
    """
    # If you want to check matrix shapes as well
    assert np.array_equal(A.shape, B.shape)
    r1, c1, v1 = sp.sparse.find(A)
    r2, c2, v2 = sp.sparse.find(B)
    np.testing.assert_equal(r1, r2)
    np.testing.assert_equal(c1, c2)
    np.testing.assert_allclose(v1, v2, atol=atol, rtol=rtol)


def get_SPD_sparse_n11_example(seed: int) -> sp.sparse.csc_array:
    """
    Create a symmetric positive definite matrix of shape(11, 11).

    Parameters
    ----------
    seed: int
        Random seed.
    """
    N = 11
    rows = np.array([5, 6, 2, 7, 9, 10, 5, 9, 7, 10, 8, 9, 10, 9, 10, 10])
    cols = np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 7, 7, 9])
    rng = np.random.default_rng(seed)
    vals = rng.random(len(rows), dtype=np.float64)
    L = sp.sparse.coo_array((vals, (rows, cols)), shape=(N, N))
    A = L + L.T  # make it symmetric
    A.setdiag(2.0)  # make it strongly positive definite
    A = A.tocsc()
    return A


def get_SPD_sparse_example(
    n: int, seed: int, diag_mean: float = 10.0
) -> sp.sparse.csc_array:
    """
    Create a symmetric positive definite matrix.

    Parameters
    ----------
    n : int
        Number of elements.
    seed : int
        Random seed.

    Returns
    -------
    sp.sparse.cscarray
        CSC sparse SPD.
    """
    # Initialize a sparse matrix
    A = sp.sparse.lil_matrix((n, n))

    # Fill the diagonal with variances (non-zero values)
    np.random.seed(42)  # for reproducibility
    variances = np.random.uniform(-1.5, 1.5, n) + diag_mean
    for i in range(n):
        A[i, i] = variances[i]

    # Add some random non-zero covariances
    for i in range(n):
        for j in range(i + 1, n):
            if np.random.rand() < 0.1:  # 10% chance of non-zero covariance
                cov = np.random.uniform(-0.5, 0.5)
                A[i, j] = cov
                A[j, i] = cov  # symmetry
    return A.tocsc()
